#
#   Imandra Inc.
#
#   __init__.py
#

from ..cl_server import CLServer
from .metamodel import register_metamodel_endpoints
from .model import register_model_endpoints
from .server import register_server_endpoints
from .sketches import register_sketches_endpoints
from .strategy import register_strategy_endpoints


def register_endpoints(app: CLServer):
    """
    Register the endpoints
    """

    register_server_endpoints(app)
    register_metamodel_endpoints(app)
    register_model_endpoints(app)
    register_sketches_endpoints(app)
    register_strategy_endpoints(app)



from datetime import datetime
from typing import List
from .models import Task, User, Project, Priority
from .storage import InMemoryStorage


class TaskService:
    def __init__(self, storage: InMemoryStorage):
        self.storage = storage

    def create_task(
        self, task_id: int, title: str, desc: str,
        priority: Priority, deadline: datetime,
        project_id: int, user_id: int | None = None
    ) -> Task:
        user = self.storage.users.get(user_id) if user_id else None
        task = Task(
            id=task_id,
            title=title,
            description=desc,
            priority=priority,
            deadline=deadline,
            assigned_to=user,
        )
        self.storage.add_task(task, project_id)
        return task

    def mark_completed(self, task_id: int):
        if task_id not in self.storage.tasks:
            raise ValueError("Task not found")
        self.storage.tasks[task_id].completed = True

    def overdue_tasks(self) -> List[Task]:
        now = datetime.now()
        return [
            t for t in self.storage.tasks.values()
            if not t.completed and t.deadline < now
        ]

    def tasks_for_user(self, user_id: int) -> List[Task]:
        return [t for t in self.storage.tasks.values()
                if t.assigned_to and t.assigned_to.id == user_id]

    def high_priority_tasks(self) -> List[Task]:
        return [t for t in self.storage.tasks.values()
                if t.priority == Priority.HIGH and not t.completed]



#
#   Imandra Inc.
#
#   config.py
#

# TODO we should just make this declarative (i.e. specify field, their types,
# default values, etc) and have the rest of the code be auto-generated. Too
# much code duplication here

import json
import logging
from enum import Enum

import yaml
from pydantic import BaseModel
from rich.panel import Panel
from rich.pretty import Pretty

log = logging.getLogger(__name__)


class StratMode(Enum):
    """
    Strategy has two modes:
    - `STANDBY` - it always waits for commands to kick off formalization and do so only in
        specific batches (i.e. )
    - `AUTO` - this continuously checks whether there's something that can be formalized
        and attempts to submit it
    """

    STANDBY = 'standby'
    AUTO = 'auto'


class StratConfigUpdate(BaseModel):
    """
    We use this to update strategies configuration
    """

    time_for_filesync_checback: int | None = None
    write_models: bool | None = None
    write_consolidated: bool | None = None
    artifact_dir: str | None = None
    iml_dir: str | None = None
    mode: StratMode | None = None
    save_on_close: bool | None = None
    meta_caching: bool | None = None


class StratConfig(BaseModel):
    """
    Container for the PyIML Strategy configuration
    """

    # fmt: off
    time_for_filesync_checkback : int = 5   # How long to wait (in seconds) for changes to local files to subside
    write_models : bool = True              # Should we write individual models (when there's iml code)
    write_models_in_same_dir : bool = True  # When we write out IML models, should we write them next to the source files?
    write_consolidated : bool = False       # Should we write a consolidated model when possible?
    artifact_dir : str = 'cldata'           # Directory with generated artifacts (relative to the source directory)
    iml_dir : str = 'iml'                   # Directory (relative to artifact directory) where IML models will be written
    mode : StratMode = StratMode.AUTO       # AUTO/STANDBY mode (this deals with formalization)
    save_on_close : bool = True             # Should we save the state/config on shutdown
    meta_caching : bool = True              # Should we be caching the meta state?
    # fmt: off

    def setField(self, field, value):
        """
        Attempt to set the specified 'field' to 'value'
        """

        def convert_str_to_bool(v: str):
            if v.lower() in ['t', 'true']:
                return True
            if v.lower() in ['f', 'false']:
                return False
            raise Exception(f"Unknown value {v} - can't convert to a bool!")

        # TODO Again, we should auto-generate this from description
        # of all the fields and their types

        if field == 'time_for_filesync_checkback':
            self.time_for_filesync_checkback = int(value)
        elif field == 'write_models':
            self.write_models = convert_str_to_bool(value)
        elif field == 'write_models_in_same_dir':
            self.write_models_in_same_dir = bool(value)
        elif field == 'write_consolidated':
            self.write_consolidated = convert_str_to_bool(value)
        elif field == 'iml_dir':
            self.iml_dir = str(value)
        elif field == 'mode':
            try:
                self.mode = StratMode[value.upper()]
            except Exception as e:
                raise Exception(
                    f"Invalid enum value for `mode` [{str(e)}]. Possible values are: 'STANDBY', 'AUTO'"
                )
        elif field == 'save_on_close':
            self.save_on_close = convert_str_to_bool(value)
        elif field == 'meta_caching':
            self.meta_caching = convert_str_to_bool(value)
        else:
            raise KeyError(f'Specified field {field} not found!')

    def __repr__(self):
        """Return a nice representation of the Config object"""
        return json.dumps(self.toJSON(), indent=4)

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Return an object from JSON file
        """

        if isinstance(j, str):
            return StratConfig.model_validate_json(j)
        elif isinstance(j, dict):
            return StratConfig.model_validate(j)
        else:
            raise Exception('Input must be either a str or a dict!')

    def toJSON(self):
        """Return a JSON object"""
        return self.model_dump_json()

    @staticmethod
    def fromYAML(path):
        """fromYAML"""

        try:
            with open(path) as infile:
                config = yaml.safe_load(infile)
        except FileNotFoundError:
            log.error(f'No file found: {path}')
            raise Exception(f'No file found: {path}')
        except yaml.YAMLError as e:
            log.error(f'Failed to parse the YAML file: {str(e)}')
            raise Exception(f'Failed to parse the YAML file: {str(e)}')
        except Exception as e:
            log.error(f'Failed to read in the YAML configuration: {str(e)}')
            raise Exception(f'Error parsing config: {str(e)}')

        # We should have a nice config container now
        try:
            config = StratConfig.model_validate(config)
        except Exception as e:
            errMsg = f'PyIMLConfig validation error: {str(e)}'
            log.error(errMsg)
            raise Exception(errMsg)

        return config

    def __rich__(self):
        """
        Return a Rich representation
        """

        pretty = Pretty(self.toJSON(), indent_size=2)
        return Panel(pretty, title='PyIMLStrategy config', border_style='green')



#
#   Imandra Inc.
#
#   codelogician/commands/test_cmd_sample.py
#

from __future__ import annotations

import json
import shutil
from pathlib import Path
from typing import Annotated, Any

import typer


def _emit_json(payload: dict[str, Any]) -> None:
    # JSON should be the only stdout output in --json mode
    typer.echo(json.dumps(payload, indent=2, sort_keys=True))


def run_sample_cmd(
    tgt_dir: Annotated[
        str, typer.Argument(help='Target directory where to place the sample project')
    ],
    json_flag: Annotated[
        bool,
        typer.Option(
            '--json',
            help='Emit machine-readable JSON to stdout (no rich output).',
        ),
    ] = False,
) -> None:
    """
    Create a sample project.
    """

    full_tgt_dir = Path(tgt_dir).expanduser().resolve()

    here = Path(__file__).resolve().parent  # .../src/codelogician/commands
    repo_root = here.parents[2]  # .../<repo-root>
    source_dir = (repo_root / 'data' / 'sample_bank_app').resolve()

    if not source_dir.exists():
        payload = {
            'ok': False,
            'error': 'SAMPLE_DATA_NOT_FOUND',
            'source_dir': str(source_dir),
            'target_dir': str(full_tgt_dir),
        }
        if json_flag:
            _emit_json(payload)
            raise typer.Exit(1)
        typer.secho(f'🛑 Sample data not found: {source_dir}', err=True)
        raise typer.Exit(1)

    if full_tgt_dir.exists():
        payload = {
            'ok': False,
            'error': 'TARGET_ALREADY_EXISTS',
            'source_dir': str(source_dir),
            'target_dir': str(full_tgt_dir),
        }
        if json_flag:
            _emit_json(payload)
            raise typer.Exit(2)
        typer.secho(f'⚠️ Target directory already exists: {full_tgt_dir}', err=True)
        raise typer.Exit(2)

    try:
        shutil.copytree(source_dir, full_tgt_dir)

        # Optional: useful metadata for callers
        copied_files = [
            str(p.relative_to(full_tgt_dir))
            for p in full_tgt_dir.rglob('*')
            if p.is_file()
        ]

        payload = {
            'ok': True,
            'source_dir': str(source_dir),
            'target_dir': str(full_tgt_dir),
            'files_copied': len(copied_files),
            'file_list': copied_files,
        }

        if json_flag:
            _emit_json(payload)
        else:
            typer.secho(f'✅ Created sample project in: {full_tgt_dir}')

    except Exception as e:
        payload = {
            'ok': False,
            'error': 'COPY_FAILED',
            'message': str(e),
            'source_dir': str(source_dir),
            'target_dir': str(full_tgt_dir),
        }
        if json_flag:
            _emit_json(payload)
        else:
            typer.secho(
                f'🛑 Failed to create sample project in {full_tgt_dir}: {e}',
                err=True,
            )
        raise typer.Exit(1)



#
#   Imandra Inc.
#
#   main.py - Main entrypoint for CodeLogician
#

import logging
from typing import Annotated, Final

import typer

from codelogician.commands.cmd_agent import run_agent_cmd
from codelogician.commands.cmd_doc import app as doc_app
from codelogician.commands.cmd_eval import app as eval_app
from codelogician.commands.cmd_multiagent import run_multiagent_cmd
from codelogician.commands.cmd_sample import run_sample_cmd
from codelogician.commands.cmd_server import app as server_app

logging.getLogger('httpx').setLevel(logging.WARNING)
logging.getLogger('httpcore').setLevel(logging.WARNING)
logging.getLogger('fsevents').setLevel(logging.WARNING)

log = logging.getLogger(__name__)


CL_HELP: Final = """
[bold deep_sky_blue1][italic]CodeLogician[/italic] is the neurosymbolic agentic governance framework for AI-powered coding.[/bold deep_sky_blue1] :rocket:

[not dim][bold italic purple]It helps your coding agent think logically about the code it's producing and test cases it's generating.[/bold italic purple]
The fundamental flaw that all LLM-powered assistants have is the reasoning they're capable of is based on statistics,
while you need rigorous logic-based automated reasoning.

- Generated code is based on explainable logic, not pure statistics
- Generated test cases are generated come with quantitative coverage metrics
- Generated code is consistent with the best security practices leveraging formal verification

To run [bold italic deep_sky_blue1]CodeLogician[/bold italic deep_sky_blue1], please obtain an [bold italic deep_sky_blue1]Imandra Universe API[/bold italic deep_sky_blue1] key available
(there's a free starting plan) at [bold italic deep_sky_blue1]https://universe.imandra.ai[/bold italic deep_sky_blue1] and make
sure it's available in your environment as [bold italic deep_sky_blue1]`IMANDRA_UNI_KEY`[/bold italic deep_sky_blue1].

[bold deep_sky_blue1 italic]Three typical workflows[/bold deep_sky_blue1 italic]:
1. [bold deep_sky_blue1 italic]DIY mode[/bold deep_sky_blue1 italic] - this is where your agent (e.g. Grok) uses the CLI to:
  - Learn how to use IML/ImandraX via `doc` command (e.g. `codelogician doc --help`)
  - Synthesizes IML code and uses the `eval` command to evaluate it
  - If there're errors, use `codelogician doc view errors` command to study how to correct the errors and re-evalute the code

2. [bold deep_sky_blue1 italic]Agent/multi-agent mode[/bold deep_sky_blue1 italic] - CodeLogician Agent is a Langgraph-based agent for automatically formalizing source code.
  - With `agent` command you can formalize a single source code file (e.g. `codelogician agent PATH_TO_FILE`)
  - With `multiagent` command you can formalize a whole directory (e.g. `codelogician agent PATH_TO_DIR`)

3. [bold deep_sky_blue1 italic]Server[/bold deep_sky_blue1 italic] - this is a "live" and interactive version of the `multiagent` command, but one you can interact with and one that
"listens" to live updates and automatically updates formalization as necessary. You can start the server and connect to it
with the TUI (we recommend separate terminal screens).

Learn more at[/not dim] [bold italic deep_sky_blue1]https://www.codelogician.dev![/bold italic deep_sky_blue1]
"""

app = typer.Typer(name='CodeLogician', help=CL_HELP, rich_markup_mode='rich')

# fmt: off
app.add_typer(server_app    , name='server'   , help='Commands for interacting with the server')
@app.command(name='tui', help='Run the TUI')
def run_tui_cmd(
    addr: Annotated[
        str, typer.Option(help='Server host/port')
    ] = 'http://127.0.0.1:8000',
):
    from codelogician.ui.main import run_tui
    run_tui(addr)
app.add_typer(doc_app       , name='doc'      , help='Documentation, guides and ImandraX reference')
app.add_typer(eval_app      , name='eval'     , help='Evaluate IML file via ImandraX API')
app.command(
    name='agent',
    help='Command for invoking CodeLogician Agent on a single file'
)(run_agent_cmd)
app.command(
    name='sample',
    help='Creates a sample Python project in specified directory (relative to the current directory)',
)(run_sample_cmd)
app.command(
    name='multiagent',
    help='Run the CL server in autoformalization mode for a directory and then quit.',
)(run_multiagent_cmd)

# fmt: on


def run_codelogician():
    app()


if __name__ == '__main__':
    app()



#
#   Imandra Inc.
#
#   codelogician/commands/server/cmd_model.py
#

from __future__ import annotations

import json
import urllib.parse
from typing import Any

import httpx
import typer

app = typer.Typer(name='model', help='CLI for CodeLogician model endpoints.')


# -------------------------
# Helpers
# -------------------------


def _print_json(data: Any) -> None:
    typer.echo(json.dumps(data, indent=2, ensure_ascii=False))


def _handle_http_error(resp: httpx.Response) -> None:
    try:
        payload = resp.json()
    except Exception:
        payload = None

    if isinstance(payload, dict) and 'detail' in payload:
        raise typer.BadParameter(f'HTTP {resp.status_code}: {payload["detail"]}')
    raise typer.BadParameter(f'HTTP {resp.status_code}: {resp.text}')


def _get(
    base_url: str,
    path: str,
    *,
    params: dict[str, Any] | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.get(url, params=params)
    if resp.status_code >= 400:
        _handle_http_error(resp)
    return resp.json()


def _post(
    base_url: str,
    path: str,
    *,
    json_body: Any | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.post(url, json=json_body)
    if resp.status_code >= 400:
        _handle_http_error(resp)

    ctype = resp.headers.get('content-type', '')
    if 'application/json' in ctype:
        return resp.json()
    return resp.text


def _cfg(ctx: typer.Context) -> dict[str, Any]:
    root = ctx.find_root()
    assert isinstance(root.obj, dict)
    return root.obj


# -------------------------
# Read endpoints
# -------------------------


@app.command('paths')
def model_paths(ctx: typer.Context) -> None:
    """GET /model/paths — list all model relative paths."""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/model/paths', timeout=cfg['timeout'])

    if cfg['json_out']:
        _print_json(data)
    else:
        for p in data:
            typer.echo(p)


@app.command('get')
def model_get(
    ctx: typer.Context,
    index: int | None = typer.Option(None, '--index', help='Model index (0-based).'),
    path: str | None = typer.Option(None, '--path', help='Model relative path.'),
) -> None:
    """
    GET a model by index or path.

    - GET /model/byindex/{index}
    - GET /model/bypath/{path}
    """
    cfg = _cfg(ctx)

    if (index is None) == (path is None):
        raise typer.BadParameter('Specify exactly one of --index or --path')

    if index is not None:
        raw = _get(cfg['base_url'], f'/model/byindex/{index}', timeout=cfg['timeout'])
        # Endpoint may return JSON (dict) or JSON-encoded string; normalize to python object.
        if isinstance(raw, str):
            try:
                data: Any = json.loads(raw)
            except Exception:
                typer.echo(raw)
                return
        else:
            data = raw
    else:
        enc_path = urllib.parse.quote(path or '', safe='')
        data = _get(
            cfg['base_url'], f'/model/bypath/{enc_path}', timeout=cfg['timeout']
        )

    if cfg['json_out']:
        _print_json(data)
    else:
        typer.echo(data)


# -------------------------
# Command endpoints
# -------------------------


@app.command('freeze')
def model_freeze(
    ctx: typer.Context,
    index: int = typer.Argument(..., help='Model index (0-based).'),
) -> None:
    """POST /model/cmd/freeze/{index} — freeze a model's IML code."""
    cfg = _cfg(ctx)
    out = _post(cfg['base_url'], f'/model/cmd/freeze/{index}', timeout=cfg['timeout'])
    typer.echo(out if isinstance(out, str) else json.dumps(out))


@app.command('unfreeze')
def model_unfreeze(
    ctx: typer.Context,
    index: int = typer.Argument(..., help='Model index (0-based).'),
) -> None:
    """POST /model/cmd/unfreeze/{index} — unfreeze a model's IML code."""
    cfg = _cfg(ctx)
    out = _post(cfg['base_url'], f'/model/cmd/unfreeze/{index}', timeout=cfg['timeout'])
    typer.echo(out if isinstance(out, str) else json.dumps(out))


@app.command('cmd')
def model_command(
    ctx: typer.Context,
    rel_path: str = typer.Argument(
        ..., help='Model relative path (as returned by `paths`).'
    ),
    cmd_json: str = typer.Option(
        ...,
        '--cmd-json',
        help='JSON object for the Command payload (sent as request body).',
    ),
) -> None:
    """
    POST /model/cmd/{path} — submit a CodeLogician agent command for a model.

    Example:
      --cmd-json '{'kind':'AutoFormalize'}'
    """
    cfg = _cfg(ctx)

    try:
        cmd_obj = json.loads(cmd_json)
    except json.JSONDecodeError as e:
        raise typer.BadParameter(f'--cmd-json must be valid JSON: {e}') from e

    enc_path = urllib.parse.quote(rel_path, safe='')
    out = _post(
        cfg['base_url'],
        f'/model/cmd/{enc_path}',
        json_body=cmd_obj,
        timeout=cfg['timeout'],
    )
    typer.echo(out if isinstance(out, str) else json.dumps(out, indent=2))


if __name__ == '__main__':
    app()



#
#   Imandra Inc.
#
#   state.py
#

import copy
import json
import logging
import os
import uuid
from pathlib import Path
from typing import Annotated

from pydantic import AfterValidator, BaseModel, Field
from rich.panel import Panel
from rich.pretty import Pretty

from ..modeling.metacache import MetaCache
from ..modeling.metamodel import MetaModel
from ..modeling.task import ModelTask
from ..server.events import ServerEvent
from ..sketch.container import SketchContainer
from ..sketch.sketch import Sketch
from ..util import json_read
from .cl_agent_state import CLResult
from .events import StrategyEvent

log = logging.getLogger(__name__)


def default_factory(data):
    return (
        data['meta_cache'].latest_mmodel()
        if data['meta_cache'].latest_mmodel()
        else MetaModel(src_dir_abs_path=data.get('src_dir_abs_path', ''), models={})
    )


class StrategyState(BaseModel):
    """
    Contains the metamodel, sketches and various additional strategy-specific settings
    """

    # Source directory - it must be absolute path!
    src_dir_abs_path: Annotated[str, AfterValidator(lambda x: str(Path(x).resolve()))]

    language: str = 'Python'

    # Container with associated sketches
    sketches: SketchContainer = SketchContainer()

    # List of outstanding CL agent tasks
    tasks: list[ModelTask] = Field(default_factory=list[ModelTask])

    # Container with a list of metamodel states along with events
    meta_cache: MetaCache = MetaCache()

    # We'll use this unique ID to keep the strategies
    strat_id: str = Field(default_factory=lambda: str(uuid.uuid4())[:8])

    #    def meta_model_validator (data: Optional[Dict[str, Any]]):
    #        if data["meta_cache"].latest():
    #            return data["meta_cache"].latest()
    #        else:
    #            return MetaModel(src_dir_abs_path=data["src_dir_abs_path"], models={})

    # Current MetaModel
    curr_meta_model: MetaModel = Field(default_factory=default_factory)

    def clear_task_ids(self):
        """
        If we're loading the state from cache, we should remove all the outstanding task IDs
        """
        for m in self.curr_meta_model.models.values():
            m.outstanding_task_ID = None

    def sketch_ids(self):
        """
        Return list of all sketch IDs
        """

        return self.sketches.ids()

    def list_sketches(self):
        """
        Return list of sketch summary information
        """
        return self.sketches.list_sketches()

    def create_new_sketch(self, anchor_model_path: str):
        """ """
        if anchor_model_path not in self.curr_meta_model.models:
            raise Exception(
                f'Could not find specified anchro model path: {anchor_model_path}'
            )

        iml_model, base_models = self.curr_meta_model.gen_consolidated_model(
            anchor_model_path
        )

        new_sketch = Sketch(
            anchor_model_path=anchor_model_path,
            init_iml_model=iml_model,
            base_models=base_models,
        )

        self.sketches.add(new_sketch)

        return new_sketch.sketch_id

    def run_file_sync(self):
        """
        Perform analysis of file system synchronization. The models's `src_code`
        are then adapted to the many things that are different.
        """

        try:
            self.curr_meta_model.run_file_sync()
        except Exception as e:
            log.error(f'Failed to perform file system synchronization: {str(e)}')
            return

    def apply_cl_result(self, cl_result: CLResult):
        """
        Apply result from running CodeLogician on a model
        """

        if self.curr_meta_model:
            self.curr_meta_model.apply_cl_result(cl_result)

    def cache_indices(self):
        """
        Return cache indicies that are available
        """
        return self.meta_cache.indices()

    def get_next_tasks(self) -> list[ModelTask]:
        """
        Return MetaModel's current tasks
        """

        # if self.curr_meta_model is None: return []

        return self.curr_meta_model.get_next_tasks()

    def save_meta_to_cache(
        self, event: ServerEvent | StrategyEvent, metaModel: MetaModel | None = None
    ):
        """
        Save the MetaModel to cache...
        """

        if metaModel is None:
            self.meta_cache.save_meta_model(event, self.curr_meta_model)
        else:
            self.meta_cache.save_meta_model(event, metaModel)

    def update_metamodel(
        self, event: ServerEvent | StrategyEvent, new_metamodel: MetaModel
    ) -> None:
        """
        Update the current MetaModel and save a snapshot to cache
        """

        self.meta_cache.save_meta_model(event, new_metamodel)
        self.curr_meta_model = new_metamodel

    def add_task(self, task: ModelTask):
        """
        Add CL task that's being worked on
        """

        self.tasks.append(task)

    def remove_task(self, task: ModelTask):
        """
        Remove CL task from the task list after it's been done
        """
        self.tasks.remove(task)

    def summary(self):
        """
        Return summary JSON of the state
        """

        currMetaModelSummary = copy.copy(self.curr_meta_model.summary())
        numModelsInCache = len(self.meta_cache.indices())

        return {
            'src_dir': self.src_dir_abs_path,
            'num_tasks': len(self.tasks),
            'current_tasks': list(map(lambda x: x.toJSON(), self.tasks)),
            'num_models_in_cache': numModelsInCache,
            'curr_meta_model': currMetaModelSummary,
        }

    @staticmethod
    def from_file(cache_file_path: str):
        """
        Load in cache from specific filepath
        """

        if not os.path.exists(cache_file_path):
            raise Exception(f'Path [{cache_file_path}] does not exist!')

        try:
            d = json_read(cache_file_path)
        except Exception as e:
            log.error(str(e))
            raise Exception(
                f'Failed to read in cache from `{cache_file_path}`: {str(e)}'
            )

        try:
            return StrategyState.fromJSON(d)
        except Exception as e:
            raise Exception(
                f'Got an error during validation: {str(e)}: path={cache_file_path}'
            )

    @staticmethod
    def from_directory(dir_path: str):
        """
        Load in the cache from a specified directory
        """

        cacheFilePath = os.path.join(dir_path, '.cl_cache')

        if os.path.exists(cacheFilePath):
            log.info('Creating strategy state from existing path')
            return StrategyState.from_file(cacheFilePath)
        else:
            log.info(
                'Specified path for strategy cache does not exist, so creating a default one'
            )
            return StrategyState(src_dir_abs_path=dir_path)

    def toJSON(self):
        """
        Return a nice JSON representation
        """
        return self.model_dump_json(indent=2)

    def save(self):
        """
        Save this state to disk
        """

        path = None
        try:
            path = os.path.join(self.src_dir_abs_path, '.cl_cache')
            with open(path, 'w') as outfile:
                print(self.toJSON(), file=outfile)
        except Exception as e:
            log.error(f'Failed to save Strategy state to: {path}; error={e}')

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Return PyIMLStrategy state object from a JSON
        """

        context = {'relink_from_str_edges': True}
        if isinstance(j, str):
            return StrategyState.model_validate_json(j, context=context)
        elif isinstance(j, dict):
            return StrategyState.model_validate(j, context=context)
        else:
            raise Exception('Input must be either a str or a dict!')

    def __repr__(self):
        """
        Should have a Rich representation also...
        """
        return f'StrategyState: \n {str(self.curr_meta_model)} '

    def __rich__(self):
        """
        Return Rich representation of the strategy state
        """

        pretty = Pretty(self.summary(), indent_size=2)
        return Panel(pretty, title='PyIML Strategy State', border_style='green')

    def eq_key(self) -> dict:
        """
        A cycle-safe, semantic equality key.
        Must NEVER include object references.
        """
        return {
            'src_dir_abs_path': self.src_dir_abs_path,
            'language': self.language,
            'strat_id': self.strat_id,
            'models': sorted(
                (
                    m.rel_path,
                    m.src_code,
                    m.agent_state.iml_model if m.agent_state else None,
                )
                for m in self.curr_meta_model.models.values()
            ),
            'dep_cycles': sorted(
                tuple(c.rel_paths)
                for c in getattr(self.curr_meta_model, 'dep_cycles', [])
            ),
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StrategyState):
            return NotImplemented
        return self.eq_key() == other.eq_key()

    def __hash__(self) -> int:
        # Only if you need hashability
        return hash(json.dumps(self.eq_key(), sort_keys=True))



#
# Imandra Inc.
#
# endpts_server.py
#

import logging

from ..cl_server import CLServer

log = logging.getLogger(__name__)


def register_server_endpoints(app: CLServer):
    """
    General admin endpoints for the server
    """

    @app.get('/server/status', operation_id='get_server_status')
    async def status():
        """
        Returns status of the server.
        """
        return {'message': "Hello, World! I'm the CodeLogician Server!"}

    @app.get('/server/tutorial', operation_id='server_tutorial')
    async def tutorial():
        """
        Returns text description of how the server works and the main functions. This is
        useful for MCP clients to learn about how to use the server itself.
        """

        tutorial = """
CodeLogician server is an


How CodeLogician works:
- It creates a formal model of the source code, while processing each file - one at a time.
- It then

Possible workflows:
- Creating a formal model of the project (directory)

        """
        return {'tutorial': tutorial}

    @app.get('/server/config', operation_id='get_server_config')
    async def server_config():
        """
        Return current server configuration
        """
        log.info('Received request for server configuration')
        return app._state.config.toJSON()



#
#   Imandra Inc.
#
#   proto_to_dict.py
#

from typing import Any

from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message


def proto_to_dict(proto_obj: Message) -> dict[Any, Any]:
    return MessageToDict(
        proto_obj,
        preserving_proto_field_name=True,
        always_print_fields_with_no_presence=True,
    )



#
#   Imandra Inc.
#
#   iml_utils.py
#

import copy
import re
from collections import deque

# ----------------------
# Regex & Helpers
# ----------------------
# fmt: off
DEF_RE = re.compile(r"^(let|type|val|verify)\s*([a-zA-Z0-9_']*)", re.MULTILINE)
ID_RE = re.compile(r"[a-zA-Z_][a-zA-Z0-9_']*")
KEYWORDS = {'let', 'type', 'val', 'in', 'match', 'fun', 'if', 'then', 'else', 'with', 'rec', 'and', 'verify'}
# fmt: on


def extract_definitions(source: str):
    """Return dict { name: full_definition_text }"""
    defs = {}
    lines = source.splitlines()
    current_name = None
    current_buf = []

    def flush():
        nonlocal current_name, current_buf
        if current_name and current_buf:
            defs[current_name] = '\n'.join(current_buf).strip() + '\n'
        current_name, current_buf = None, []

    for line in lines:
        m = DEF_RE.match(line.strip())
        if m:
            flush()
            current_name = m.group(2)
            current_buf = [line]
        elif current_buf is not None:
            current_buf.append(line)
    flush()
    return defs


def free_vars(def_text: str):
    tokens = set(ID_RE.findall(def_text))
    return tokens - KEYWORDS


def topo_sort_definitions(defs: dict):
    """
    Topologically sort definitions based on dependencies on other defs only.
    """
    # Build dependency graph: node -> set of nodes it depends on (only other defs)
    graph = {
        name: ((free_vars(body) & set(defs.keys())) - set([name]))
        for name, body in defs.items()
    }

    # Compute in-degree
    in_degree = {name: len(deps) for name, deps in graph.items()}

    queue = deque([name for name, deg in in_degree.items() if deg == 0])
    sorted_list = []

    while queue:
        n = queue.popleft()
        sorted_list.append(n)

        # Remove n from dependencies of other nodes
        for m in defs:
            if n in graph[m]:
                graph[m].remove(n)
                in_degree[m] -= 1
                if in_degree[m] == 0:
                    queue.append(m)

    if len(sorted_list) != len(defs):
        # Optional: print graph for debugging
        for name, deps in graph.items():
            if deps:
                print(f'{name} still depends on {deps}')
        raise RuntimeError('Cycle detected in definitions, cannot sort topologically.')

    return sorted_list


# ----------------------
# Operations
# ----------------------


def add_definition(source: str, new_def: str) -> str:
    defs = extract_definitions(source)
    m = DEF_RE.search(new_def)

    is_verify = False

    if not m:
        raise ValueError('New definition must start with let/type/val or verify')

    kind, new_name = m.groups()
    if kind == 'verify':
        is_verify = True
        # for verify, generate a pseudo-name if empty
        if not new_name:
            new_name = f'__verify_{hash(new_def) & 0xFFFF}'

    defs = extract_definitions(source)

    if new_name in defs:
        raise ValueError(f'Definition {new_name} already exists')

    combined = source.strip() + '\n' + new_def.strip() + '\n'
    new_defs = extract_definitions(combined)

    if is_verify:
        new_defs[new_name] = new_def

    sorted_defs = topo_sort_definitions(new_defs)

    return ''.join(list(new_defs[d] for d in sorted_defs))
    # return source.strip() + "\n" + new_def.strip() + "\n"


def update_definition(
    source: str, name: str, new_def: str, verify_index: int = 0
) -> str:
    """
    Updates a definition (let/type/val) or a verify statement in the source IML code.

    - For let/type/val, 'name' must be the identifier.
    - For verify statements, set name="verify" and use verify_index to choose which one to update.
    """
    # Handle verify statements
    if name == 'verify':
        regex = re.compile(r'(verify\s*\([^\n]*?\);{0,2})', re.MULTILINE)
        matches = list(regex.finditer(source))
        if not matches:
            raise ValueError('No verify statements found to update')
        if verify_index >= len(matches):
            raise ValueError(
                f'verify_index {verify_index} out of range (found {len(matches)} verify statements)'
            )
        match = matches[verify_index]
        start, end = match.span()
        return source[:start] + new_def.strip() + source[end:]

    # Normal let/type/val updates
    pattern = rf'(let|type|val)\s+{re.escape(name)}\b[^\n]*?=(?:[\s\S]*?)(?=\n(let|type|val|verify|end|\Z))'
    regex = re.compile(pattern, re.MULTILINE)
    if regex.search(source):
        return regex.sub(new_def.strip(), source)

    raise ValueError(f'Failed to perform update: definition {name} not found')


def remove_definition(source: str, name: str, verify_index: int | None = None) -> str:
    """
    Removes a definition (let/type/val) or a verify statement from the source IML code.

    - For let/type/val, 'name' must be the identifier.
    - For verify statements, set name="verify" and optionally provide verify_index to remove a specific one.
      If verify_index is None, all verify statements will be removed.
    """
    # Handle verify statements
    if name == 'verify':
        regex = re.compile(r'(verify\s*\([^\n]*?\);{0,2})', re.MULTILINE)
        matches = list(regex.finditer(source))
        if not matches:
            raise ValueError('No verify statements found to remove')

        if verify_index is None:
            # Remove all verify statements
            return regex.sub('', source)
        else:
            if verify_index >= len(matches):
                raise ValueError(
                    f'verify_index {verify_index} out of range (found {len(matches)} verify statements)'
                )
            match = matches[verify_index]
            start, end = match.span()
            return source[:start] + source[end:]

    # Normal let/type/val removal
    pattern = rf'\n?(let|type|val)\s+{re.escape(name)}\b[^\n]*?=(?:[\s\S]*?)(?=\n(let|type|val|verify|end|\Z))'
    regex = re.compile(pattern, re.MULTILINE)
    if regex.search(source):
        return regex.sub('', source)

    raise ValueError(f'Failed to perform remove: definition {name} not found')


def sync_definitions_topo(
    combined: str,
    originals: dict[str, str],
    misc_file_name: str = 'misc.iml',
):
    originals = copy.deepcopy(originals)

    combined_defs = extract_definitions(combined)

    known_defs: dict[str, str] = {}
    orig_names: set[str] = set()

    # -------------------------
    # 1) Update existing defs in-place (by name)
    # -------------------------
    for orig_file_path, src in list(originals.items()):
        orig_defs = extract_definitions(src)

        updated_src = src

        # IMPORTANT: make update_definition's regex happy by ensuring newline termination
        if updated_src.strip() and not updated_src.endswith('\n'):
            updated_src = updated_src + '\n'

        # record what exists
        for name, body in orig_defs.items():
            known_defs[name] = body
            orig_names.add(name)

        # update any changed defs using regex-based update_definition
        for name, old_body in orig_defs.items():
            if name not in combined_defs:
                continue

            new_body = combined_defs[name]
            if new_body.strip() != old_body.strip():
                updated_src = update_definition(
                    source=updated_src,
                    name=name,
                    new_def=new_body.strip(),
                )
                # keep bookkeeping consistent with what we applied
                known_defs[name] = new_body

        # normalize file formatting back to your convention
        originals[orig_file_path] = (
            updated_src.rstrip() + '\n' if updated_src.strip() else ''
        )

    # -------------------------
    # 2) New defs -> misc.iml (unchanged from your version)
    # -------------------------
    new_defs = {n: b for n, b in combined_defs.items() if n not in orig_names}

    if new_defs:
        sorted_names = topo_sort_definitions(new_defs)
        misc_path = misc_file_name
        target_text = originals.get(misc_path, '')

        all_known_defs = dict(known_defs)

        for name in sorted_names:
            body = new_defs[name]
            deps = free_vars(body) & set(all_known_defs.keys())

            lines = target_text.splitlines()  # DO NOT STRIP
            insert_index = 0
            for i, line in enumerate(lines):
                m = DEF_RE.match(line.strip())
                if m and m.group(2) in deps:
                    insert_index = i + 1

            target_text = '\n'.join(
                lines[:insert_index]
                + ['']
                + [body.rstrip()]
                + ['']
                + lines[insert_index:]
            )
            all_known_defs[name] = body

        originals[misc_path] = (
            target_text.rstrip() + '\n' if target_text.strip() else ''
        )

    return originals



#
#   CodeLogician Server
#
#   codelogician/server/cl_server.py
#

import logging
import os
from contextlib import asynccontextmanager
from pathlib import Path

import watchdog.events
import watchdog.observers
from fastapi import FastAPI
from fastapi_mcp import FastApiMCP

from ..modeling.model import Model
from ..sketch.sketch import Sketch
from ..strategy.config import StratConfig
from ..strategy.pyiml_strategy import PyIMLStrategy
from ..strategy.state import StrategyState
from .events import FileSystemEvent, FileSystemEventType
from .file_event_handler import MyFileSysEventHandler
from .state import ServerState

log = logging.getLogger(__name__)


@asynccontextmanager
async def lifespan(app: 'CLServer'):
    """
    This handles events that happen on start/shutdown of the FastAPI server
    """

    for strat_worker in app.strategy_workers.values():
        app.file_sys_observer.schedule(
            MyFileSysEventHandler(app.file_event_callback),
            strat_worker.state().src_dir_abs_path,
            recursive=True,
        )

        if not strat_worker.is_alive():
            strat_worker.start()
            log.info(
                f'Strategy thread [{strat_worker.state().src_dir_abs_path}] started'
            )

    app.file_sys_observer.start()
    log.info('File system observer started!')

    # Let's now pass on the control to other parts of the system
    yield

    # This should kill the `run` loop
    log.info('Sending `None` event to strategies to shut them down!')
    for strategy in app.strategy_workers.values():
        strategy.add_event(None)

    # Let's send in these just in case
    app.file_sys_observer.stop()

    # Now let's wait for them to finish
    for strategy in app.strategy_workers.values():
        strategy.join()  # Let's wait for the strategy to stop
        strategy.state().save()  # Now let's save its state to disk

    app.file_sys_observer.join()

    app.save_state()


class CLServer(FastAPI):
    """Class for CL Server"""

    def __init__(self, server_state: ServerState):
        """ctor"""
        super().__init__(title='CLServer', lifespan=lifespan)
        self._state = server_state  # Config is part of the state

        self.current_strat_id: str | None = None

        # This is the container of all strategy workers (threads)
        self.strategy_workers: dict[str, PyIMLStrategy] = {}

        # If we have initial strategies in our config, we will now create
        # actual strategy threads for them
        self.create_strat_workers()

        # Observer will help us monitor filesystem events
        self.file_sys_observer = watchdog.observers.Observer()

        # Now let's setup the MCP server
        if self._state.config.mcp:
            self.mount_mcp_server()

    def set_current_strat_id(self, new_id: str):
        if new_id not in self.strategy_workers:
            raise Exception(f'Invalid strategy id: {new_id}')

    def search_sketches(self, query: str) -> list[Model]:
        """
        return list of models for specified query
        """
        return []

    def all_sketch_ids(self) -> list[str]:
        """
        Return the full list of sketch IDs
        """
        all_ids: list[str] = []
        for strat_id in self.strategy_workers:
            all_ids.extend(self.strategy_workers[strat_id].state().sketch_ids())

        return all_ids

    def get_strat_for_sketch_id(self, sketch_id: str) -> PyIMLStrategy | None:
        """
        return strategy ID for a corresponding sketch ID (or None if can't find it)
        """
        for strat_id in self.strategy_workers:
            if sketch_id in self.strategy_workers[strat_id].state().sketch_ids():
                return self.strategy_workers[strat_id]

        return None

    def get_sketch_from_sketch_id(self, sketch_id: str) -> Sketch | None:
        """
        Return Sketch object from the specified sketch_id
        """
        for strat_id in self.strategy_workers:
            if sketch_id in self.strategy_workers[strat_id].state().sketch_ids():
                return (
                    self.strategy_workers[strat_id].state().sketches.from_id(sketch_id)
                )

        return None

    def create_sketch(self, strat_id: str, anchor_model_path: str) -> str:
        """
        Create a new sketch and return its sketch ID
        """
        if strat_id not in self.strategy_workers:
            raise Exception(f'Could not locate the specified strategy: {strat_id}')

        return (
            self.strategy_workers[strat_id].state().create_new_sketch(anchor_model_path)
        )

    def list_sketches(self) -> dict[str, list[dict[str, str]]]:
        """
        Return list of sketches by strategy
        """
        return {i: s.state().list_sketches() for i, s in self.strategy_workers.items()}

    def list_strategies(self) -> list[dict[str, str]]:
        """
        Return a list of dictionaries specifying the strategies currently running
        """
        strats = []
        for strat_id in self.strategy_workers:
            strat = {
                'strat_id': strat_id,
                'path': self.strategy_workers[strat_id].state().src_dir_abs_path,
                'type': 'Python',  # for now, we only support python
            }
            strats.append(strat)

        return strats

    def create_strat_workers(self):
        """We go through the list of paths we have, initialize their states"""

        log.info(
            f'Will now instantiate server strategies. I have {len(self._state.strategy_paths)} paths!'
        )

        for path in list(set(self._state.strategy_paths)):
            try:
                strat_config = StratConfig()
                strat_state = StrategyState.from_directory(path)
                self.strategy_workers[strat_state.strat_id] = PyIMLStrategy(
                    state=strat_state, config=strat_config
                )

                if self.current_strat_id is None:
                    self.current_strat_id = strat_state.strat_id

            except Exception as e:
                log.error(f'Error during strategy instantiation: {path}: {str(e)}')
                continue

            log.info(f'Created strategy worker for {path}.')

    def setup_strat_configs(self, strat_config: str):
        """Load up configuration for strategies"""
        try:
            stratConfig: StratConfig = StratConfig.fromYAML(strat_config)
            log.info(f'Reading in StratConfig file: {strat_config}')
            log.info(stratConfig)
        except Exception as e:
            log.error(f'Failed to load StratConfig: {str(e)}. Switching to defaults.')
            stratConfig = StratConfig()

        return stratConfig

    # def setup_server_config(self, server_config:str):
    #     # Let's try to load in the server config file
    #     try:
    #         serverConfig = ServerConfig.fromYAML(server_config)
    #         log.info (f"Reading in configuration file: {server_config}")
    #         log.info (serverConfig)
    #     except Exception as e:
    #         log.error(f"Failed to load server configuration: {str(e)}. Switching to defaults.")
    #         serverConfig = ServerConfig()

    #     return serverConfig

    def mount_mcp_server(self):
        """
        Mount the MCP server
        """
        log.info('Starting to mount MCP server...')
        mcp = FastApiMCP(
            self,
            name='CodeLogician Server',
            description='MCP server for CodeLogician Server',
            describe_full_response_schema=True,
            describe_all_responses=True,
        )
        # Mount the MCP server directly to your FastAPI app
        mcp.mount_http()
        log.info('MCP server mounted!')

    def strategy_worker_by_id(self, strat_id: str) -> PyIMLStrategy:
        """
        Return strategy worker for specified ID
        """
        if strat_id not in self.strategy_workers:
            raise Exception(f'Specified ID not found: {strat_id}')

        return self.strategy_workers[strat_id]

    def strategy_state_by_id(self, strat_id: str) -> StrategyState:
        """
        Return strategy state for specified ID
        """
        if strat_id not in self.strategy_workers:
            raise Exception(f'Specified ID not found: {strat_id}')

        return self.strategy_workers[strat_id].state()

    def strategy_states(self) -> dict[str, StrategyState]:
        """
        Return dictionary of all strategy states
        """
        return {s.state().strat_id: s.state() for s in self.strategy_workers.values()}

    def strategy_config_by_id(self, strat_id: str) -> StratConfig:
        """
        Return strategy config for specified ID
        """
        if strat_id not in self.strategy_workers:
            raise Exception(f'Specified ID not found: {strat_id}')

        return self.strategy_workers[strat_id].config()

    def update_strat_config(self, strat_id: str, strat_config: StratConfig):
        """
        Update strategy configuration for the specified strategy
        """
        if strat_id not in self.strategy_workers:
            raise Exception(f'Specified strategy [{strat_id}] not found!')

        self.strategy_workers[strat_id].update_config(strat_config)

    @staticmethod
    def is_path_valid(new_path: str, existing_paths: list[str]) -> bool:
        """
        Return True/False if the specified path doesn't overlap with existing strategies' paths
        """
        if new_path in existing_paths:
            return False

        for curr_path in existing_paths:
            if Path(curr_path).is_relative_to(Path(new_path)) or Path(
                new_path
            ).is_relative_to(Path(curr_path)):
                return False

        return True

    def add_strategy(self, strat_type: str, new_strat_path: str) -> str:
        """
        Add a new strategy and return its ID
        """
        new_strat_path = str(Path(new_strat_path).resolve())

        if not os.path.exists(new_strat_path):
            raise Exception(
                f"Failed to create new strategy: specified path [{new_strat_path}] doesn't exist!"
            )

        existing_paths = [
            s.state().src_dir_abs_path for s in self.strategy_workers.values()
        ]

        if not self.is_path_valid(new_strat_path, existing_paths):
            raise Exception(f'Specified path is not valid: {new_strat_path}')

        if Path(new_strat_path).is_dir():
            try:
                strat_state = StrategyState.from_directory(new_strat_path)
            except Exception:
                strat_state = StrategyState(src_dir_abs_path=new_strat_path)
        else:
            try:
                strat_state = StrategyState.from_file(new_strat_path)
            except Exception:
                strat_state = StrategyState(src_dir_abs_path=new_strat_path)

        _ = strat_type  # FIXME: strat_type currenty ignored

        strat_config = StratConfig()
        try:
            self.strategy_workers[strat_state.strat_id] = PyIMLStrategy(
                state=strat_state, config=strat_config
            )
            self.strategy_workers[strat_state.strat_id].start()
        except Exception as e:
            raise Exception(f'Failed to create ther worker: {e}')

        self.file_sys_observer.schedule(
            MyFileSysEventHandler(self.file_event_callback),
            self.strategy_workers[strat_state.strat_id].state().src_dir_abs_path,
            recursive=True,
        )

        log.info(f'Started Strategy for dir: {strat_state.src_dir_abs_path}')

        if self.current_strat_id is None:
            self.current_strat_id = strat_state.strat_id

        return strat_state.strat_id

    def rem_strategy(self, strat_id: str) -> None:
        """
        Remove strategy (by strat_id)
        """
        if strat_id not in self.strategy_workers:
            raise Exception(f'Specified strategy path does not exist: {strat_id}')

        self.strategy_workers[strat_id].add_event(None)  # this should stop it
        self.strategy_workers[strat_id].join()

        if strat_id == self.current_strat_id:
            # let's try to find another strat_id instead of this one
            self.current_strat_id = next(
                (x for x in self.strategy_workers if x != strat_id), None
            )

        del self.strategy_workers[strat_id]

    def path_to_strategy(self, filepath: str) -> str | None:
        """
        Return strategy ID if the file path is contained within this
        """

        def good_extension(extensions):
            for ext in extensions:
                if filepath.endswith(ext):
                    return True
            return False

        for strat_id in self.strategy_workers:
            strat_state: StrategyState = self.strategy_workers[strat_id].state()

            if Path(filepath).is_relative_to(Path(strat_state.src_dir_abs_path)):
                # Let's check that this is not a directory and the right file extension for this strategy
                if not Path(filepath).is_dir() and good_extension(
                    self.strategy_workers[strat_id].extensions()
                ):
                    return strat_id

        return None

    def file_event_callback(
        self, event: watchdog.events.FileSystemEvent, event_type: FileSystemEventType
    ) -> None:
        """
        Gets called by the FileSystemEventObserver - this routes event to the right strategy
        """
        src_path = os.fsdecode(event.src_path)
        strat_id = self.path_to_strategy(src_path)

        if strat_id is None:
            log.info(
                f'Received an event, but couldnt assign it to a strategy: {event.src_path}; {event_type}'
            )
            return

        strategy = self.strategy_workers[strat_id]

        if event_type in [
            FileSystemEventType.CREATED,
            FileSystemEventType.MODIFIED,
            FileSystemEventType.DELETED,
        ]:
            ext = Path(src_path).suffix
            if ext in strategy.extensions():
                self.strategy_workers[strat_id].add_event(
                    FileSystemEvent(action_type=event_type, abs_path1=src_path)
                )
            else:
                log.warning(
                    f'Received a filsystem event for file with unsupported extension: {ext}'
                )

        elif event_type == FileSystemEventType.MOVED:
            src_ext = Path(src_path).suffix
            dest_path = os.fsdecode(event.dest_path)
            dest_ext = Path(dest_path).suffix

            if src_ext in strategy.extensions() and dest_ext in strategy.extensions():
                self.strategy_workers[strat_id].add_event(
                    FileSystemEvent(
                        action_type=event_type,
                        abs_path1=src_path,
                        abs_path2=dest_path,
                    )
                )
            else:
                log.warning(
                    f'Received a filesystem event for file with unsupported extension: {src_ext} -> {dest_ext}'
                )

        else:
            log.error(f'Received an unknown event type: {event_type}')

    def save_state(self) -> None:
        """
        Attempt to save the current server state information. Note that we're only recording the
        paths to strategies - they will save their own states there...
        """
        log.info('Will save server state now.')
        self._state.strategy_paths = []

        for s in self.strategy_workers.values():
            log.info(f'Adding {str(s.state().src_dir_abs_path)}')
            self._state.strategy_paths.append(str(s.state().src_dir_abs_path))

        try:
            self._state.save()
        except Exception as e:
            log.error(f'Failed to save server state to disk: {e}')



"""
CLI that only includes the eval and doc subcommands.

For dev-purposes mainly.
"""
#
#   Imandra Inc.
#
#   tool_cli.py
#

import typer

from codelogician.commands.cmd_doc import app as doc_app  # noqa: F401
from codelogician.commands.cmd_eval import app as eval_app

app = typer.Typer(name='CodeLogician', no_args_is_help=True)

# app.add_typer(doc_app, name='doc', help='Documentation, guides and ImandraX reference')
app.add_typer(eval_app, name='eval', help='Evaluate IML file via ImandraX API')


if __name__ == '__main__':
    app()



#
#   Imandra Inc.
#
#   codelogician/modeling/renderer.py
#

import json

from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.pretty import Pretty
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text

from ..strategy.cl_agent_state import frm_status_to_rich
from ..util import intersperse, maybe_else
from .dep_context import ModelDepContext
from .metamodel import MetaModel
from .model import Model
from .task_creator import ModelTaskCreator

# ------------------------------------------------------------------------------------------
#   Model renderers
# ------------------------------------------------------------------------------------------


def model_to_json(m: Model):
    # This will save the reasons which we can only calculate if
    # we have "live" dependency models (not strings as is the case
    # post-deserialization)

    # Removed this from Model class
    # @field_serializer('static_frm_reasons')
    # def serialize_frm_reasons(self, _):
    #    return self.formalization_reasons()

    task_creator = ModelTaskCreator()
    m.static_frm_reasons = task_creator.formalization_reasons(m, 5)

    return m.model_dump_json()


def model_to_str_summary(m: Model) -> str:
    return str(ModelTaskCreator().gen_model_status(m))


def model_to_str(m: Model) -> str:
    agentStateStr = 'None' if m.agent_state is None else m.agent_state.status

    s = f'Model: {m.rel_path} \n'
    # s += f'{str(ModelTaskCreator().gen_model_status.status())}\n'
    s += f'Formalization state: {agentStateStr}\n'
    s += f'Depends on: {ModelDepContext.deps_paths(m)}\n'
    s += f'Source language: {m.src_language}\n'

    if m.src_code:
        s += f'Source code (condensed): \n {m.src_code[:100]}\n'
    else:
        s += 'Source code (condensed): \n None \n'

    s += f'Opaque funcs: {len(m.opaque_funcs())}\n'
    s += f'Decomps: {len(m.decomps())}\n'
    return s


def iml_code_to_print(m: Model) -> str:
    """
    Provide IML code (if available) along with relevant import statements
    """
    if m.iml_code() is None:
        return 'N\\A'

    import_statements = ModelDepContext().create_import_iml_statements(m)
    return f'{import_statements}\n\n{m.iml_code()}'


def render_model(model: Model):
    table = Table(
        title=Text(f'Model: {model.rel_path}', style='bold italic magenta'),
        box=box.MINIMAL_DOUBLE_HEAD,
        show_lines=True,
    )

    table.add_column('Name')
    table.add_column('Attribute')

    src_code_str = 'N/A' if model.src_code is None else model.src_code
    user_iml_code_str = (
        'N/A' if model.user_iml_edit is None else model.user_iml_edit.user_iml_entry
    )
    user_iml_code_last_change = (
        'N/A' if model.user_iml_edit is None else str(model.user_iml_edit.time_of_edit)
    )

    needs_frm_str = 'Yes' if model.static_frm_reasons == [] else 'No'
    frozen_str = 'Yes' if model.iml_code_frozen else 'No'

    iml_code_str = 'N/A' if not model.iml_code() else str(model.iml_code())

    eval_result_str = 'N/A' if not model.agent_state else model.agent_state.eval_res

    # fmt: off
    table.add_row('Relative path',              Text(model.rel_path, style='bold')                      )
    table.add_row('Formalization status',       frm_status_to_rich(model.formalization_status())        )
    table.add_row('Eval result',                Text(eval_result_str, style='bold')                    )
    table.add_row('Needs formalization?',       Text(needs_frm_str, style='bold')                      )
    table.add_row('Formalization reasons',      Text(str(model.static_frm_reasons))                     )
    table.add_row('Frozen?',                    Text(frozen_str)                                       )
    table.add_row('Src last changed',           Text(f'{model.src_code_last_changed}', style='bold')    )
    table.add_row('Src code',                   Syntax(src_code_str, 'python')                         )
    table.add_row('IML code',                   Syntax(iml_code_str, 'ocaml')                          )
    table.add_row('User IML code',              Syntax(user_iml_code_str, 'ocaml')                     )
    table.add_row('User IML code last changed', Text(f'{user_iml_code_last_change}', style='bold')     )
    table.add_row('Context',                    Text(f'{model.context}', style='bold')                  )
    table.add_row('Task ID',                    Text(f'{model.outstanding_task_ID}', style='bold')      )
    table.add_row('Deps',                       Text(f'{model.dependencies}', style='bold')             )
    table.add_row('Opaques',                    Text(f'{model.opaque_funcs()}', style='bold')           )
    table.add_row('Verification Goals',         Text(f'{model.verification_goals()}', style='bold')     )
    table.add_row('Decompositions',             Text(f'{model.decomps()}', style='bold')                )
    # fmt: on

    return table


def render_model_list(models: list[Model]):
    table = Table(
        title=Text('Models list', style='bold italic magenta'),
        box=box.MINIMAL_DOUBLE_HEAD,
        show_lines=True,
    )

    # fmt: off
    table.add_column('ID'                       , justify='right', no_wrap=True )
    table.add_column('Frm Status'               , justify='right', no_wrap=True )
    table.add_column('Frozen'                   , justify='right', no_wrap=True )
    table.add_column('Needs frm?'               , justify='right', no_wrap=True )
    table.add_column('Frm Needs'                , justify='right', no_wrap=False)
    table.add_column('Rel path'                 , justify='right', no_wrap=True )
    table.add_column('Src Last Changed'         , justify='right', no_wrap=True )
    table.add_column('User IML Code'            , justify='right', no_wrap=True )
    table.add_column('User IML Code Last Chged' , justify='right', no_wrap=True )
    table.add_column('Src Code Embds'           , justify='right', no_wrap=True )
    table.add_column('IML Code Embds'           , justify='right', no_wrap=True )
    table.add_column('Context'                  , justify='right' )
    table.add_column('Task ID'                  , justify='right' )
    table.add_column('Dependencies'             , justify='right', no_wrap=False)
    table.add_column('Num opaques'              , justify='right' )
    table.add_column('Num VGs'                  , justify='right' )
    table.add_column('Num Decomps'              , justify='right' )

    def yes_no(b): return 'Yes' if b else 'No'
    def edit_time(e): return str(e.time_of_edit)
    # fmt: on

    for m_id, m in enumerate(models):
        table.add_row(
            Text(f'{m_id}', style='bold'),
            frm_status_to_rich(m.formalization_status()),
            Text(f'{yes_no(m.iml_code_frozen)}', style='bold'),
            Text(f'{yes_no(m.static_frm_reasons)}', style='bold'),
            Text(f'{map(str, m.static_frm_reasons)}', style='bold'),
            Text(f'{m.rel_path}', style='bold'),
            Text(f'{m.src_code_last_changed}', style='bold'),
            Text(f'{m.user_iml_edit is not None}', style='bold'),
            Text(f'{maybe_else("N/A", edit_time, m.user_iml_edit)}', style='bold'),
            Text(f'{m.context}', style='bold'),
            Text(f'{m.outstanding_task_ID}', style='bold'),
            Text(f'{m.dependencies}', style='bold'),
            Text(f'{len(m.opaque_funcs())}', style='bold'),
            Text(f'{len(m.verification_goals())}', style='bold'),
            Text(f'{len(m.decomps())}', style='bold'),
        )

    return table


# ------------------------------------------------------------------------------------------
#   MetaModel renderers
# ------------------------------------------------------------------------------------------


def metamodel_rich_summary(mmodel: MetaModel):
    """Return rich summary of the metamodel state"""
    summary = mmodel.summary()

    def frm_status(status_count):
        status, count = status_count
        return Text.assemble(status.__rich__(), ' ', (str(count), 'bold'))

    def fmt_countable(label_key):
        label, key = label_key
        return Text.assemble(f'Num. {label}: ', (str(summary[key]), 'bold'))

    countables = [
        ('Models', 'num_models'),
        ('Verification goals', 'num_vgs'),
        ('Opaque functions', 'num_opaque_funcs'),
        ('Decompositions', 'num_decomps'),
    ]
    return Group(
        Text.assemble(*intersperse(' | ', map(fmt_countable, countables))),
        Text.assemble(
            *intersperse(' | ', map(frm_status, summary['frm_statuses'].items()))
        ),
    )


def mmodel_to_str(mmodel: MetaModel):
    """Return a nice representation"""

    metaSummary = mmodel.summary()
    statuses = {s.name: n for s, n in metaSummary['frm_statuses'].items()}

    s = f'Source file directory: {mmodel.src_dir_abs_path}\n'
    s += f'Number of models: {metaSummary["num_models"]}\n'
    s += f'Number of VGs: {metaSummary["num_vgs"]}\n'
    s += f'Number of decomps: {metaSummary["num_decomps"]}\n'
    s += f'Status breakdown: \n{json.dumps(statuses, indent=4)}'
    s += '\n'
    s += 'Model details:\n'

    for path in mmodel.models:
        s += f'Model [path={path}; {model_to_str_summary(mmodel.models[path])}]\n'

    return s


def __rich__(self):
    """Return RICH values"""
    pretty = Pretty(self.toJSON(), indent_size=2)
    return Panel(pretty, title='Current MetaModel', border_style='green')



#
#   Imandra Inc.
#
#   iml_strategy.py
#

import logging
from threading import Thread

from ..server.events import FileSystemEvent
from .state import StrategyState

log = logging.getLogger(__name__)


class IMLStrategy(Thread):
    """
    Base class for all IML (ImandraX) strategies
    """

    def __init__(self, state: StrategyState):
        super().__init__()
        self._state = state

    def watch_directories(self):
        """Return the list of directories the observer should watch"""
        return []

    def on_load(self):
        """What should be done on startup"""
        pass

    def on_filesystem_event(self, event: FileSystemEvent):
        """ """
        log.info(f'Received {event} to process')

    def on_save(self):
        """
        What should be done on strategy save
        """
        pass



from typing import Dict
from .models import Account, Transaction


class InMemoryStorage:
    def __init__(self):
        self.accounts: Dict[int, Account] = {}
        self.transactions: Dict[int, Transaction] = {}

    def add_account(self, account: Account):
        self.accounts[account.id] = account

    def add_transaction(self, txn: Transaction):
        self.transactions[txn.id] = txn



#
#   Imandra Inc.
#
#   tree_views.py
#

import os.path
from typing import Generic, TypeVar

from rich.style import Style
from rich.text import Text
from textual import on
from textual.containers import Container
from textual.message import Message
from textual.reactive import reactive
from textual.widgets import TabbedContent, TabPane, Tree
from textual.widgets.tree import TreeNode

from ..modeling.metamodel import MetaModel
from ..modeling.task import TaskKind
from ..modeling.task_creator import ModelTaskCreator
from ..util import fst, maybe, maybe_else, mdict, snd, splice

TOGGLE_STYLE = Style.from_meta({'toggle': True})
T = TypeVar('T')

ok, unk, inad = '●', '○', '●'  # "⨂", "✓", "-", "x"
status_indicators = {
    'unknown': (unk, 'dim'),
    'transparent': (ok, 'light_green'),
    'inadmissible': (inad, 'bright_red'),
    'admitted_with_opaqueness': (ok, 'bright_cyan'),
    'executable_with_approximation': (ok, 'bright_yellow'),
    'in_progress': ('•', 'blink bold bright_white'),
}


class GraphView(Tree):
    class MoveRequest(Generic[T], Message):
        def __init__(self, node: TreeNode[T]) -> None:
            self.node: TreeNode[T] = node
            super().__init__()

    # fmt: off
    def __init__(
        self, root_label, id, node_info, get_successors, get_descendants, level1, prev_views
    ):
        Tree.__init__(self, root_label, id=id)
        node_by_path = {}
        prev_view = prev_views.get(id)
        prev_root = maybe(lambda t: t.root, prev_view)
        prev_selected_src_node = maybe(lambda t:maybe(lambda n: maybe(fst, n.data), t.cursor_node), prev_view)

        def add_children(prev_node, node, xs):
            def node_sort_key(node): return fst(fst(node_info(node)))
            def node_children_by_label(node):
                return {fst(n.data): n for n in node.children if n.data is not None}
            prev_children = maybe_else({}, node_children_by_label, prev_node)
            for x in sorted(xs, key=node_sort_key):
                add_subtree(prev_children, node, x)

        def add_subtree(prev_children, dest, src_node):
            successors = get_successors(src_node)
            (label, get_fresult), path = node_info(src_node)
            add_node = dest.add_leaf if successors == [] else dest.add
            node = add_node(label, data=(src_node, get_fresult))
            prev_node = prev_children.get(src_node)
            if prev_node is not None and prev_node.is_expanded:
                node.expand()
            if path:
                node_by_path[path] = node

            add_children(prev_node, node, successors)
            if prev_selected_src_node == src_node:
                self.post_message(GraphView.MoveRequest(node))

        self.path_of_node = lambda n: snd(node_info(n))
        self.node_by_path = node_by_path
        self.get_descendants = get_descendants
        self.root.expand()
        add_children(prev_root, self.root, level1)

    def render_label(self, node, base_style, style):
        # if self.mmodel is None:
        #     return Text()
        node_label = node._label.copy()
        node_label.stylize(style)
        def indicators_of_st(status, analyzing):
            return [status_indicators.get(status, ('?', 'red')),
                    (status_indicators['in_progress'] if analyzing else (' ', ''))]
        def text(sym_col): return Text(fst(sym_col), snd(sym_col))
        def toggled_icon():
            icon = self.ICON_NODE_EXPANDED if node.is_expanded else self.ICON_NODE
            return icon, base_style + TOGGLE_STYLE

        sym_cols = maybe_else([], splice(indicators_of_st), maybe(lambda d: snd(d)(), node.data))
        prefix = toggled_icon() if node._allow_expand else ('', base_style)
        return Text.assemble(prefix, *map(text, sym_cols), node_label)
    # fmt: off

    @on(MoveRequest)
    def move_to(self, event):
        self.move_cursor(event.node)

        # if fullpath in self.mmodel.models:
        #     model = self.mmodel.models[fullpath]
        # else:
        #     model = None


def fs_graph(metamodel):
    import os.path

    def nodes_of_path(is_file, path):
        yield path, is_file
        if path != '':
            yield from nodes_of_path(False, os.path.dirname(path))

    def edges_of_path(path):
        if path != '':
            parent = os.path.dirname(path)
            yield parent, path
            yield from edges_of_path(parent)

    paths = list(metamodel.models.keys())
    nodes = {n: is_file for p in paths for n, is_file in nodes_of_path(True, p)}
    edges = {e for p in paths for e in edges_of_path(p)}

    succ = mdict(edges)
    return nodes, lambda n: succ.get(n, [])


def dep_graph(metamodel):
    from codelogician.util import consult_or, distrib, map, map_values, swap

    def module_deps(m):
        return [mm.rel_path for mm in m.dependencies]

    deps = map_values(module_deps, metamodel.models)
    preds = mdict(map(swap, distrib(deps.items())))
    inits = [n for n, _ in deps.items() if not preds.get(n)]
    terms = [n for n, succs in deps.items() if not succs]
    succ = consult_or([], {'': inits, **deps})
    pred = consult_or([], {'': terms, **preds})
    return inits, terms, succ, pred


class TreeViews(Container):
    IDS = {'src-tree', 'module-deps', 'rev-deps'}
    DEFAULT_CSS = 'TreeViews { width: 35% }'
    mmodel = reactive(MetaModel(src_dir_abs_path='<not set>', models={}))

    def __init__(self):
        self.prev_view = None
        Container.__init__(self)

    def active_tree(self):
        active_pane = self.query_one('#tree-views', TabbedContent).active_pane
        assert active_pane is not None
        return active_pane.query_one(GraphView)

    def watch_mmodel(self, old, new):
        self.call_next(self.diff_recompose, old, new)

    async def diff_recompose(self, _old, new):
        from textual.compose import compose

        # return await self.recompose()
        async with self.batch():
            graph_views = {w.id: w for w in self.query(GraphView)}
            children = self.query_children('*').exclude('.-textual-system')
            try:
                prev_view = self.query_one(TabbedContent).active, graph_views
            except Exception:
                prev_view = None
            await children.remove()
            if self.is_attached:
                self.prev_view = prev_view
                await self.mount_all(compose(self))
                self.prev_view = None

    def compose(self):
        if self.mmodel is not None:
            fs_nodes, fs_succs = fs_graph(self.mmodel)
            dg_inits, dg_terms, dg_succs, dg_preds = dep_graph(self.mmodel)
            model_by_path = self.mmodel.models
            prev_active_id, prev_views = (
                ('', {}) if self.prev_view is None else self.prev_view
            )

            task_creator = ModelTaskCreator()

            def result_of_path(path):
                status = task_creator.gen_model_status(model_by_path[path])

                # fmt: off
                def task_is(t): return status.outstanding_task_kind == t
                formalization_state = (
                    'in_progress' if task_is(TaskKind.FORMALIZE)
                    else status.formalization_status.value
                )
                # fmt: on
                return formalization_state, task_is(TaskKind.ANALYZE)

            # transitive closure
            def tc(f):
                def g(x):
                    return {z for y in f(x) for z in [y, *g(y)]}

                return g

            # fmt: off
            def fs_node_info(rel_path):
                path = rel_path if fs_nodes[rel_path] else None
                return (os.path.basename(rel_path), lambda: maybe(result_of_path, path)), path

            def dg_node_info(rel_path):
                return (rel_path, lambda: result_of_path(rel_path)), rel_path

            with TabbedContent(id='tree-views', initial=prev_active_id):
                with TabPane('Source tree'):
                    root = os.path.basename(self.mmodel.src_dir_abs_path)
                    yield GraphView(root, 'src-tree', fs_node_info, fs_succs, tc(fs_succs), fs_succs(''), prev_views)
                with TabPane('Module deps'):
                    yield GraphView('⊤', 'module-deps', dg_node_info, dg_succs, tc(dg_succs), dg_inits, prev_views)
                with TabPane('Rev deps'):
                    yield GraphView('⊥', 'rev-deps', dg_node_info, dg_preds, tc(dg_preds), dg_terms, prev_views)
            # fmt: on

    def on_tree_node_selected(self, event):
        def select_path(rel_path):
            def expand_and_select(node):
                def expand_up(n):
                    n.expand()
                    maybe(expand_up, n.parent)

                maybe(expand_up, node.parent)
                other.post_message(GraphView.MoveRequest(node))

            for other_id in self.IDS - {event.control.id}:
                other = self.query_one('#' + other_id, GraphView)
                maybe(expand_and_select, other.node_by_path.get(rel_path))

        def select_node(node_data):
            maybe(select_path, event.control.path_of_node(fst(node_data)))

        maybe(select_node, event.node.data)



#
#   Imandra Inc.
#
#   codelogician/commands/server/cmd_sketches.py
#

from __future__ import annotations

import json
import urllib.parse
from typing import Any

import httpx
import typer

app = typer.Typer(name='sketches', help='CLI for CodeLogician sketches endpoints.')


# -------------------------
# Helpers
# -------------------------


def _print_json(data: Any) -> None:
    typer.echo(json.dumps(data, indent=2, ensure_ascii=False))


def _handle_http_error(resp: httpx.Response) -> None:
    try:
        payload = resp.json()
    except Exception:
        payload = None

    if isinstance(payload, dict) and 'detail' in payload:
        raise typer.BadParameter(f'HTTP {resp.status_code}: {payload["detail"]}')
    raise typer.BadParameter(f'HTTP {resp.status_code}: {resp.text}')


def _get(
    base_url: str,
    path: str,
    *,
    params: dict[str, Any] | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.get(url, params=params)
    if resp.status_code >= 400:
        _handle_http_error(resp)
    return resp.json()


def _post(
    base_url: str,
    path: str,
    *,
    params: dict[str, Any] | None = None,
    json_body: Any | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.post(url, params=params, json=json_body)
    if resp.status_code >= 400:
        _handle_http_error(resp)

    ctype = resp.headers.get('content-type', '')
    if 'application/json' in ctype:
        return resp.json()
    return resp.text


def _cfg(ctx: typer.Context) -> dict[str, Any]:
    root = ctx.find_root()
    assert isinstance(root.obj, dict)
    return root.obj


# -------------------------
# Commands
# -------------------------


@app.command('search')
def search_sketches(
    ctx: typer.Context,
    query: str = typer.Argument('', help='Search query string.'),
) -> None:
    """GET /sketches/search?query=..."""
    cfg = _cfg(ctx)
    data = _get(
        cfg['base_url'],
        '/sketches/search',
        params={'query': query},
        timeout=cfg['timeout'],
    )

    if cfg['json_out']:
        _print_json(data)
    else:
        if isinstance(data, list):
            for item in data:
                typer.echo(str(item))
        else:
            typer.echo(str(data))


@app.command('list')
def list_sketches(ctx: typer.Context) -> None:
    """GET /sketches/list"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/sketches/list', timeout=cfg['timeout'])

    if cfg['json_out']:
        _print_json(data)
        return

    # Expected: dict[str, list[dict[str, str]]]
    if isinstance(data, dict):
        for strat_id, items in data.items():
            typer.echo(f'Strategy {strat_id}:')
            if isinstance(items, list):
                for it in items:
                    if isinstance(it, dict):
                        sid = it.get('sketch_id', '<unknown>')
                        anchor = it.get(
                            'anchor_model', it.get('anchormodel', '<unknown>')
                        )
                        typer.echo(f'  - {sid} (anchor: {anchor})')
                    else:
                        typer.echo(f'  - {it}')
            else:
                typer.echo(f'  {items}')
    else:
        typer.echo(str(data))


@app.command('create')
def create_sketch(
    ctx: typer.Context,
    anchor_model_path: str = typer.Argument(..., help='Anchor model path.'),
) -> None:
    """
    POST /sketches/create

    Note: endpoint signature is `create_sketch(anchor_model_path: str)` with no explicit
    Body/Query annotation. In FastAPI, this is typically treated as a query parameter.
    """
    cfg = _cfg(ctx)
    data = _post(
        cfg['base_url'],
        '/sketches/create',
        params={'anchor_model_path': anchor_model_path},
        timeout=cfg['timeout'],
    )
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('try-change')
def try_change(
    ctx: typer.Context,
    sketch_id: str = typer.Argument(..., help='Sketch ID.'),
    change_json: str = typer.Option(
        ...,
        '--change-json',
        help='JSON payload for SketchChange (request body).',
    ),
    commit_on_success: bool = typer.Option(
        True,
        '--commit/--no-commit',
        help='Whether to commit the change automatically on success (default: commit).',
    ),
) -> None:
    """
    POST /sketches/{sketch_id}/try_change?commit_on_success=...

    Body: SketchChange JSON.
    """
    cfg = _cfg(ctx)

    try:
        change_obj = json.loads(change_json)
    except json.JSONDecodeError as e:
        raise typer.BadParameter(f'--change-json must be valid JSON: {e}') from e

    sid = urllib.parse.quote(sketch_id, safe='')
    data = _post(
        cfg['base_url'],
        f'/sketches/{sid}/try_change',
        params={'commit_on_success': commit_on_success},
        json_body=change_obj,
        timeout=cfg['timeout'],
    )
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('change')
def apply_change(
    ctx: typer.Context,
    sketch_id: str = typer.Argument(..., help='Sketch ID.'),
    change_json: str = typer.Option(
        ...,
        '--change-json',
        help='JSON payload for one of SketchChgSetModel | SketchChgInsertDef | SketchChgModifyDef | SketchChgDeleteDef.',
    ),
) -> None:
    """POST /sketches/{sketch_id}/change (body: SketchChg*)"""
    cfg = _cfg(ctx)

    try:
        change_obj = json.loads(change_json)
    except json.JSONDecodeError as e:
        raise typer.BadParameter(f'--change-json must be valid JSON: {e}') from e

    sid = urllib.parse.quote(sketch_id, safe='')
    data = _post(
        cfg['base_url'],
        f'/sketches/{sid}/change',
        json_body=change_obj,
        timeout=cfg['timeout'],
    )
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('rollback')
def rollback(
    ctx: typer.Context,
    sketch_id: str = typer.Argument(..., help='Sketch ID.'),
    target_state_id: int | None = typer.Option(
        None,
        '--target-state-id',
        help='Rollback to a specific state id. If omitted, roll back the last change.',
    ),
) -> None:
    """POST /sketches/{sketch_id}/rollback?target_state_id=..."""
    cfg = _cfg(ctx)

    sid = urllib.parse.quote(sketch_id, safe='')
    params = (
        {'target_state_id': target_state_id} if target_state_id is not None else None
    )
    data = _post(
        cfg['base_url'],
        f'/sketches/{sid}/rollback',
        params=params,
        timeout=cfg['timeout'],
    )
    _print_json(data) if cfg['json_out'] else typer.echo(str(data))


@app.command('state')
def latest_state(
    ctx: typer.Context,
    sketch_id: str = typer.Argument(..., help='Sketch ID.'),
) -> None:
    """
    POST /sketches/{sketch_id}/state

    Note: current server endpoint returns 'OK' (placeholder).
    """
    cfg = _cfg(ctx)

    sid = urllib.parse.quote(sketch_id, safe='')
    data = _post(cfg['base_url'], f'/sketches/{sid}/state', timeout=cfg['timeout'])
    if cfg['json_out'] and isinstance(data, (dict, list)):
        _print_json(data)
    else:
        typer.echo(str(data))


if __name__ == '__main__':
    app()



#
#  Imandra Inc.
#
#  codelogician/strategy/worker.py
#

import asyncio
import logging
import os
from collections.abc import Callable
from functools import partial
from queue import Queue
from threading import Thread
from typing import Any

import dotenv
import imandrax_api
from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
from httpx import HTTPStatusError
from imandra.core import AsyncClient
from imandra.u.agents import create_thread_sync, get_remote_graph
from imandra.u.agents.code_logician.base import FormalizationStatus
from imandra.u.agents.code_logician.base.region_decomp import (
    DecomposeReqData,
    RegionDecomp,
)
from imandra.u.agents.code_logician.base.vg import VG, VerifyReqData
from imandra.u.agents.code_logician.graph import GraphState
from imandrax_api_models import DecomposeRes, VerifyRes
from iml_query.processing import extract_decomp_reqs, extract_verify_reqs
from iml_query.tree_sitter_utils import get_parser

from codelogician.util import fst, get_imandra_uni_key, maybe, maybe_else

from ..modeling.task import ModelTask, UserManualIMLEditTask
from ..sketch.task import SketchChangeResult, SketchChangeTask
from .cl_agent_state import CLAgentState, CLResult, CLResultStatus
from .events import CLResultEvent, SketchChangeResultEvent


def ensure_dir(dir):
    if not os.path.exists(dir):
        os.makedirs(dir)


def exc_to_none(msg, f):
    try:
        return f()
    except Exception as e:
        log.error(f'{msg}: {repr(e)}')
        return None


def proto_to_dict(proto_obj: Message) -> dict[Any, Any]:
    return MessageToDict(
        proto_obj,
        preserving_proto_field_name=True,
        always_print_fields_with_no_presence=True,
    )


dotenv.load_dotenv('.env')

log = logging.getLogger(__name__)


def printer_callback(result: CLResult | SketchChangeResult):
    """
    Simple printer callback function to use in testing
    """

    if isinstance(result, CLResult):
        print(f'The CLResult is: {result}')
    elif isinstance(result, SketchChangeResult):
        print(f'The ImandraXResult is {result} ')
    else:
        print(f'Unknown task type: {type(result).__name__}')


# def proto_to_dict(proto_obj: Message) -> dict:
#    """imandrax-api returns protobuf messages, this function converts them to
#    dictionaries"""
#    return MessageToDict(
#        proto_obj,
#        preserving_proto_field_name=True,
#        always_print_fields_with_no_presence=True,
#    )


async def run_sketch_task(
    task: SketchChangeTask,
    callback: Callable[[SketchChangeResultEvent], None] | None = None,
) -> SketchChangeResult | None:
    """
    Run ImandraX to analyze the sketch change. If `callback` function is provided, then call it with the result.
    """

    try:
        iml = task.iml_code
        tree = get_parser().parse(iml.encode('utf-8'))
        _iml, _tree, verify_reqs, _ranges = extract_verify_reqs(iml, tree)
        # _iml, _tree, instance_reqs, _ranges = extract_instance_reqs(iml, tree)
        _iml, _tree, decomp_reqs, _ranges = extract_decomp_reqs(iml, tree)
    except Exception as e:
        raise Exception(f'Error during extraction of artifacts from IML code: {e}')

    decomp_results, verify_results = [], []

    try:
        async with AsyncClient(
            url=imandrax_api.url_prod,
            auth_token=get_imandra_uni_key(),
        ) as c:
            eval_res = await c.eval_src(task.iml_code)  # pyright: ignore

            # We only run these if we could parse the file
            if eval_res.success:
                # instance_results = [
                #    c.instance(**instance_req) for instance_req in instance_reqs
                # ]

                if len(decomp_reqs):
                    decomp_results = [
                        await c.decompose(**dr)  # pyright: ignore
                        for dr in decomp_reqs
                    ]
                else:
                    decomp_results = []

                if verify_reqs:
                    verify_results = [
                        await c.verify_src(**vr)  # pyright: ignore
                        for vr in verify_reqs
                    ]
                else:
                    verify_results = []

            # if len(verify_results + decomp_results):
            #    await asyncio.gather(*(verify_results + decomp_results))

    except Exception as e:
        raise Exception(f'Error during call to ImandraX: {e}')

    # eval_res = EvalRes.model_validate(proto_to_dict(eval_res_data))

    region_decomps, vgs = [], []
    if eval_res.success:
        # Fill region decomps
        for decomp_req, decomp_res in zip(decomp_reqs, decomp_results, strict=True):
            decomp_req_data_model = DecomposeReqData(**decomp_req)  # pyright: ignore
            decomp_res_model = DecomposeRes.model_validate(proto_to_dict(decomp_res))
            region_decomps.append(
                RegionDecomp(
                    data=decomp_req_data_model,
                    res=decomp_res_model,
                )
            )

        # Fill vgs
        for verify_req, verify_res in zip(verify_reqs, verify_results, strict=True):
            verify_req_data_model = VerifyReqData(
                predicate=verify_req['src'], kind='verify'
            )
            verify_res_model = VerifyRes.model_validate(proto_to_dict(verify_res))
            vgs.append(
                VG(
                    data=verify_req_data_model,
                    res=verify_res_model,
                )
            )

    result = SketchChangeResult(
        task=task,
        success=eval_res.success,
        error=str(eval_res.errors),
        vgs=vgs,
        decomps=region_decomps,
    )
    if callback:
        callback(
            SketchChangeResultEvent(sketch_id=task.sketch_id, change_result=result)
        )
    else:
        return result


def run_code_logician(
    task: ModelTask, callback: Callable[[CLResultEvent], None] | None
) -> CLResult | None:
    """Run CodeLogician agent on the specified task object and return the result via the callback function"""

    def process_result(res):
        def make_agent_state(f):
            import codelogician.tools.cl_iml as ci

            log.info(f'Result from CodeLogician for {task.rel_path}: {f.status}')
            errors = f.eval_res.all_errors
            sections = ci.full_iml_model(f)
            describe_error = partial(ci.describe_error, task.rel_path, sections)
            error_contexts = [maybe_else([], describe_error, e.msg) for e in errors]

            # if f.status == FormalizationStatus.INADMISSIBLE:
            #     target = f'inadmissibles/{task.rel_path}.iml'
            #     log.info(f'Dumping inadmissible result to `{target}')
            #     try:
            #         ensure_dir(os.path.dirname(target))
            #         with open(target, 'w') as fp:
            #             fp.write('\n'.join(ci.join_sections(sections)))
            #     except Exception as e:
            #         log.info(f'Dump failed: {repr(e)}')

            eval_res_obj = f.eval_res
            errors = eval_res_obj.all_errors if eval_res_obj is not None else []
            eval_res_str = str(eval_res_obj) if eval_res_obj is not None else ''

            return CLAgentState(
                status=f.status,
                src_code=f.src_code,
                iml_code=f.iml_code,
                iml_model=f.iml_model,
                vgs=f.vgs,
                region_decomps=f.region_decomps,
                opaque_funcs=f.opaque_funcs,
                graph_state=res,
                errors=errors,
                error_contexts=error_contexts,
                eval_res=eval_res_str,
            )

        return maybe(make_agent_state, res.last_fstate)

    def formalise():
        # Let's inititialize the graph
        graph = get_remote_graph('code_logician', api_key=get_imandra_uni_key())
        hint = 'Try checking that your Imandra API key is set correctly and enabled in your online Imandra Universe settings at `https://universe.imandra.ai/user/api-keys`.'
        try:
            create_thread_sync(graph)
        except HTTPStatusError as e:
            if e.response.status_code == 403:
                reason = e.response.json()['error']
                raise Exception(f'Access to Imandra API denied ("{reason}"). ${hint}')
            else:
                raise

        # If the task has an existing graph state, let's just use that as the starting point
        # Our task should contain the commands we need to add
        gs = (task.graph_state or GraphState()).add_commands(task.commands())

        def run():
            return fst(asyncio.run(gs.run(graph)))

        typestr = (
            'UserManualIMLEditTask'
            if isinstance(task, UserManualIMLEditTask)
            else 'ModelTask'
        )
        log.info(f'Sending task to CodeLogician[{typestr}]: {task.rel_path}')
        return maybe(
            process_result, exc_to_none(f'CodeLogician failed for {task.rel_path}', run)
        )

    def formalise_empty():
        return CLAgentState(
            status=FormalizationStatus.TRANSPARENT,
            src_code='',
            iml_code='',
            iml_model='',
        )

    def process_agent_state(agent_state):
        r = CLResult(task=task, status=CLResultStatus.SUCCESS, agent_state=agent_state)

        def apply(callback):
            return exc_to_none(
                'Exception executing callback',
                lambda: callback(CLResultEvent(result=r)),
            )

        return maybe_else(r, apply, callback)

    agent_state = formalise_empty() if task.is_empty() else formalise()
    return maybe(process_agent_state, agent_state)


class CodeLogicianWorker(Thread):
    """Thread for processing CL requests"""

    def __init__(self, callback=None):
        """ """
        super().__init__(daemon=True)

        self._callback = callback
        self._queue = Queue()

    def add_task(self, task: ModelTask | SketchChangeTask | None):
        """
        Add a model task to the queue
        """
        log.info('Adding task to the queue')
        try:
            self._queue.put_nowait(task)
        except Exception as e:
            log.error(f"Couldn't add to the queue: {e}")

    def run(self):
        """
        Let's execute some CL requests
        """
        while True:
            try:
                task = self._queue.get()  # Blocks until an item is available
                log.info('Processing new task.')

                if task is None:  # Sentinel value to signal termination
                    log.info("Received new task 'None'. Will now shutdown.")
                    break

                if isinstance(task, ModelTask):
                    target = run_code_logician
                elif isinstance(task, SketchChangeTask):
                    target = run_sketch_task
                else:
                    raise Exception(f'Unsupported type of task: {type(task).__name__}')

                Thread(target=target, args=(task, self._callback)).start()
                # we've fired off the request, so now it should be good to go...
                self._queue.task_done()

            except Exception:
                # This block won't be reached if get() is called without a timeout
                pass



#
#   Imandra Inc.
#
#   codelogician/modeling/mlist.py
#

from functools import cmp_to_key
from typing import cast

from pydantic import BaseModel

from .model import Model


class ModelList(BaseModel):
    models: list[Model]


def calc_upstream_affected(models: dict[str, Model], rel_path: str) -> int:
    """
    Return the number of models affected by the model with specified 'rel_path'
    """

    if rel_path not in models:
        raise Exception(f'Specified model [{rel_path}] is not present!')

    def helper(current_path: str, visited: list[str]):
        """Our helper function to help us traverse the models"""

        for path in models:
            if path in visited:
                continue

            found = False

            for d in models[path].dependencies:
                if current_path == cast(Model, d).rel_path:
                    found = True
                    break

            if found:
                visited.append(path)
                visited = helper(path, visited)

        return visited

    return len(helper(rel_path, []))


def gen_listing(models: dict[str, Model], category: str = 'frm_status') -> ModelList:
    """
    Return a sorted list of models by a specified category.

    Possible values are:
    - Formalization Status [value='frm_status'] - formalization status
    - Upstream dependencies [value='upstream'] - number of upstream dependencies
    - Number of opaque functions [value='opaques'] - number of opaque functions
    - Number of failed VGs [value='failed_vgs'] - number of failed verification goals
    """

    supported = ['frm_status', 'upstream', 'opaques', 'failed_vgs']

    # Let's make sure the category is supported first
    if category.lower() not in supported:
        raise Exception(
            f'Supplied category is not supported: {category}. Possible values are: {supported}'
        )

    stats = {}

    # This should get us the number of models that are ultimatey affected by this model
    for rel_path in models:
        upstream = calc_upstream_affected(models, rel_path)
        stats[rel_path] = {'path': rel_path, 'upstream': upstream}

    for rel_path in models:
        modelStats = models[rel_path].gen_stats()

        # Let's now append the statistics here
        for key in modelStats:
            stats[rel_path][key] = modelStats[key]

    def compare(s1, s2):
        try:
            if category.lower() == 'frm_status':
                return s1['frm_status'] >= s2['frm_status']
            elif category.lower() == 'upstream':
                return s1['upstream'] >= s2['upstream']
            elif category.lower() == 'opaques':
                return s1['num_opaques'] >= s2['num_opaques']
            else:
                return s1['failed_vgs'] >= s2['failed_vgs']
        except Exception as e:
            raise Exception(f'Failed to compare model statistcs: {str(e)}')

    return ModelList(models=list(sorted(stats.values(), key=cmp_to_key(compare))))



#
#   Imandra Inc.
#
#   app.py
#

import datetime
import json
import os
from pathlib import Path
from typing import Annotated

import httpx
import typer
from textual import work
from textual.app import App
from textual.reactive import reactive
from textual.screen import Screen

from ..modeling.metamodel import MetaModel
from ..sketch.container import SketchContainer
from ..strategy.config import StratConfigUpdate
from ..strategy.state import StrategyState
from ..util import head, maybe
from .commands import ModelCommandsProvider, ServerCommandsProvider
from .common import InfoScreen, Source, Status
from .screens.decomps import DecompsScreen
from .screens.help import HelpScreen
from .screens.intro import IntroScreen
from .screens.model import ModelScreen
from .screens.opaques import OpaquesScreen
from .screens.overview import OverviewScreen
from .screens.sketches import SketchScreen
from .screens.vgs import VGsScreen

app = typer.Typer()


# fmt: off
def curr_meta_model(s): return s.curr_meta_model
def sketches(s): return s.sketches
# fmt: on


class CodeLogicianTUI(App):
    CSS_PATH = 'tui.tcss'

    status: reactive[Status] = reactive(Status(source=Source.NoSource))
    strat_states: reactive[dict[str, StrategyState]] = reactive({})
    mmodel: reactive[None | MetaModel] = reactive(None)
    sketch_container: reactive[None | SketchContainer] = reactive(None)
    last_update: reactive[str] = reactive('---')

    # fmt: off
    MODES = {
        'intro'     : IntroScreen,
        'overview'  : OverviewScreen,
        'model'     : ModelScreen,
        'opaques'   : OpaquesScreen,
        'vgs'       : VGsScreen,
        'decomps'   : DecompsScreen,
        'sketch'    : SketchScreen,
        'help'      : HelpScreen
    }

    DEFAULT_MODE = 'intro'
    BINDINGS = [
        ('i', "switch_mode('intro')"    , 'Intro'               ),
        ('o', "switch_mode('overview')" , 'Overview'            ),
        ('m', "switch_mode('model')"    , 'Model'               ),
        ('v', "switch_mode('vgs')"      , 'Verification Goals'  ),
        ('d', "switch_mode('decomps')"  , 'Decomps'             ),
        ('k', "switch_mode('opaques')"  , 'Opaques'             ),
        ('s', "switch_mode('sketch')"   , 'Sketches'            ),
        ('h', "switch_mode('help')"     , 'Help'                ),
        ('q', 'quit'                    , 'Quit'                )
    ]
    # fmt: on

    COMMANDS = App.COMMANDS | {ServerCommandsProvider} | {ModelCommandsProvider}

    def __init__(self, status: Status):
        super().__init__()
        self.status = status

        # We should now have the issue resolved altogether
        self.console.set_window_title('Imandra CodeLogician v2.0')

    @work()
    async def cl_server_state_retriever(self):
        """
        Polls the server to retrieve the latest state information
        """

        if self.status.source != Source.Server:
            return

        status = self.status.copy()
        try:
            endpoint = f'{status.server_addr}/strategy/states'
            blob = json.loads(httpx.get(endpoint).text.strip("'"))
            self.strat_states = {
                strat_state.src_dir_abs_path: strat_state
                for strat_state in map(StrategyState.fromJSON, blob)
            }
            status.server_last_update = datetime.datetime.now()
            some_strat_state = head(self.strat_states.values())
            # TODO: only override if not set or current setting no longer exists
            self.mmodel = maybe(curr_meta_model, some_strat_state)
            self.sketch_container = maybe(sketches, some_strat_state)
            self.last_update = str(status.server_last_update)
            status.server_error = None

        except Exception as e:
            msg = f'Error when connecting to the server: {e}'
            status.server_error = msg

        self.status = status

    def on_mount(self) -> None:
        if self.status.source == Source.Disk:
            self.load_disk_strat_states(self.status.disk_path, True)  # pyright: ignore
        elif self.status.source == Source.Server:
            self.set_server(self.status.server_addr)  # pyright: ignore

    def on_unmount(self) -> None:
        """on_unmount"""
        self.cont = False

    def do_update_strat_config(self, config: StratConfigUpdate):
        """
        This is the callback for PyIMLStrategy configuration
        """
        pass

    def do_execute_server_command(self, command):
        pass

    def do_view_model(self, model_path: str):
        """
        Focus the view on a specific model
        """
        pass

    def do_select_strategy(self, strat_path: str):
        """Select a specific strategy"""
        # TODO: what if strat_path is invalid?
        self.mmodel = maybe(curr_meta_model, self.strat_states[strat_path])
        self.sketch_container = maybe(sketches, self.strat_states[strat_path])

    def set_server(self, host: str):
        """
        Set the source to be the server
        """

        self.status.source = Source.Server
        self.status.server_addr = host
        self.status.disk_error = None
        self.selected_strat = None
        self.timer = self.set_interval(5, self.cl_server_state_retriever)
        self.cl_server_state_retriever()

    def load_disk_strat_states(self, path: Path, startup: bool = False):
        """Here, we're loading the strategy state directly from disk"""

        status = self.status.copy()
        status.source = Source.Disk
        status.server_error = None
        status.server_last_update = None
        status.disk_path = path

        load = (
            StrategyState.from_directory if path.is_dir() else StrategyState.from_file
        )
        try:
            state = load(str(path))
            status.disk_error = None
            self.strat_states = {state.src_dir_abs_path: state}
            self.mmodel = state.curr_meta_model
        except Exception as e:
            errMsg = f'Error loading from disk: {str(e)}'
            if startup:
                print(errMsg)
            else:
                self.push_screen(InfoScreen(errMsg, 'error'))
            status.disk_error = errMsg

        self.status = status

    def get_system_commands(self, screen: Screen):
        yield from super().get_system_commands(screen)


@app.command()
def run_tui(
    server_addr: Annotated[
        str, typer.Option(help='Server to connect to by default')
    ] = 'http://127.0.0.1:8000',
):
    status = Status(
        source=Source.Server, server_addr=server_addr, disk_path=Path(os.getcwd())
    )

    CodeLogicianTUI(status=status).run()


@app.command()
def run_tui_disk(
    disk_path: Annotated[
        str, typer.Option(help='Path to .cl_cache file to load in disk mode')
    ],
):
    """This is useful for testing."""
    status = Status(source=Source.Disk, disk_path=Path(disk_path))
    CodeLogicianTUI(status=status).run()


if __name__ == '__main__':
    app()



from .models import Account, Transaction, Entry
from .storage import InMemoryStorage


class AccountingService:
    def __init__(self, storage: InMemoryStorage):
        self.storage = storage
        self.next_account_id = 1
        self.next_txn_id = 1

    def create_account(self, name: str, type_: str) -> Account:
        account = Account(id=self.next_account_id, name=name, type=type_)
        self.storage.add_account(account)
        self.next_account_id += 1
        return account

    def post_transaction(self, description: str, entries: list[Entry]) -> Transaction:
        txn = Transaction(id=self.next_txn_id, description=description, entries=entries)
        if not txn.is_balanced():
            raise ValueError("Transaction not balanced: debits != credits")

        # Update account balances
        for entry in entries:
            account = self.storage.accounts[entry.account_id]
            account.balance += entry.debit
            account.balance -= entry.credit

        self.storage.add_transaction(txn)
        self.next_txn_id += 1
        return txn

    def get_account_balances(self):
        return {acc.name: acc.balance for acc in self.storage.accounts.values()}

    def trial_balance(self):
        total_debits = 0.0
        total_credits = 0.0
        for acc in self.storage.accounts.values():
            if acc.balance >= 0:
                total_debits += acc.balance
            else:
                total_credits += -acc.balance
        return total_debits, total_credits



#
#   Imandra Inc.
#
#   codelogician/commands/cmd_agent.py
#

import asyncio
import json
from pathlib import Path
from typing import Annotated

import typer
from rich import print as printr


def run_agent_cmd(
    file: Annotated[
        Path, typer.Argument(help='Path to a source code file to run the CL agent on')
    ],
    outfile: Annotated[
        Path,
        typer.Argument(help='Output file path to save the resulting CL Agent state'),
    ] = Path('cl_result.json'),
    artifacts: Annotated[
        bool,
        typer.Option(
            help='Should we generate artifacts like Verification Goals and Decomposition requests'
        ),
    ] = True,
):
    """
    Read the provided file and run CodeLogician Agent on it, saving the result in `outfile` file.
    """

    from imandra.u.agents.code_logician import SUPPORTED_LANGUAGES

    from codelogician.tools.cl_caller import call_cl_agent

    if not file.exists():
        typer.secho(f'🛑 File does not exist: {file}')
        raise typer.Exit(1)

    language = None
    for lang, ext in SUPPORTED_LANGUAGES:
        if file.suffix == ext:
            language = lang
            break

    if language is None:
        typer.secho(
            f'🛑 Unsupported language with extension: {file.suffix}. Only the following are supported: {SUPPORTED_LANGUAGES}'
        )
        raise typer.Exit(1)

    try:
        src_code = file.read_text()
    except Exception as e:
        typer.secho(f'🛑 Failed to read the file: {e}')
        raise typer.Exit(1)

    try:
        res = asyncio.run(
            call_cl_agent(src_code=src_code, language=language, artifacts=artifacts)
        )
    except Exception as e:
        typer.secho(
            f'🛑 Caught exception during the call: {e}', fg=typer.colors.RED, err=True
        )
        raise typer.Exit(1)

    if res is None:
        typer.secho('🛑 Failed to execute the CL agent. Received `None`', err=True)
        raise typer.Exit(1)

    printr(res)

    try:
        with open(outfile, 'w') as outfile_file:
            data = res.model_dump(mode='json')
            fields = [
                'status',
                'src_code',
                'src_lang',
                'eval_res',
                'vgs',
                'region_decomps',
                'iml_code',
                'iml_model',
            ]
            select_data = {k: data[k] for k in fields if k in data}
            print(json.dumps(select_data, indent=4), file=outfile_file)
            typer.secho(f'✅ Written results to {outfile}!')

    except Exception as e:
        typer.secho(f'🛑 Failed to print to out file: {e}', err=True)
        raise typer.Exit(1)



#
#   Imandra Inc.
#
#   overview.py
#

import os
from collections.abc import Iterable
from pathlib import Path

from textual import on
from textual.containers import (
    HorizontalGroup,
    VerticalGroup,
    VerticalScroll,
)
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import (
    Button,
    DirectoryTree,
    Footer,
    Input,
    Select,
    Static,
    TabbedContent,
    TabPane,
)

from codelogician.modeling.metamodel import MetaModel
from codelogician.modeling.renderer import metamodel_rich_summary
from codelogician.strategy.state import StrategyState
from codelogician.util import dup, filter, map, maybe_else

from ..common import Border, MyHeader, Source, Status, bind


class FilteredDirectoryTree(DirectoryTree):
    """We just want to show directories and `cl_cache` files."""

    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        pred = lambda p: p.name.startswith('.cl_cache') or os.path.isdir(p)  # noqa
        return filter(pred, paths)


class SelectFileDirectory(Screen):
    DEFAULT_CSS = """VerticalGroup { align: center middle; width: 70%; height: 90% }"""

    def __init__(self, initial_path):  # , name = None, id = None, classes = None):
        # super().__init__(name, id, classes)
        super().__init__()
        self.styles.align_horizontal = 'center'
        self.styles.align_vertical = 'middle'
        self._path_selected = initial_path

    def compose(self):
        with VerticalGroup():
            root = self._path_selected or os.getcwd()
            yield Static('Select directory or cache file location')
            yield FilteredDirectoryTree(path=root, id='directory_tree')
            with HorizontalGroup():
                yield Button('Go up', action='screen.go_up')
                yield Button('Select', action='screen.select')
                yield Button('Cancel', action='screen.cancel')

    @on(DirectoryTree.DirectorySelected)
    def select_directory(self, event):
        self._path_selected = event.path

    def action_cancel(self):
        self.dismiss(None)

    def action_select(self):
        self.dismiss(self._path_selected)

    def action_go_up(self):
        directoryTree = self.query_one('#directory_tree', DirectoryTree)
        directoryTree.path = os.path.dirname(directoryTree.path)
        directoryTree.reload()


class DiskOptions(VerticalGroup):
    disk_path = reactive('', recompose=True)

    def __init__(self, status):
        super().__init__()
        self.disk_path = status.disk_path

    def compose(self):
        yield Static(content=f'Current selection: {self.disk_path or ""}')
        with HorizontalGroup():
            yield Button('Select source', id='directory_view')
            yield Button('Apply', id='apply_disk')

    @on(Button.Pressed, '#apply_disk')
    def on_apply_disk(self, event: Button.Pressed):
        self.app.load_disk_strat_states(self.disk_path)  # pyright: ignore[reportAttributeAccessIssue]

    @on(Button.Pressed, '#directory_view')
    def on_directory_view(self, event: Button.Pressed):
        def get_path(path: str | None):
            if path:
                self.disk_path = path

        self.app.push_screen(SelectFileDirectory(self.disk_path), get_path)


class ServerOptions(VerticalGroup):
    def __init__(self, status):
        super().__init__()
        self.status = status

    def compose(self):
        yield Static('Enter CodeLogician server details:')
        yield Input(value=self.status.server_addr or '', id='server_address')
        yield Button('Apply', id='apply_server')

    @on(Button.Pressed, '#apply_server')
    def on_apply_server(self) -> None:
        host = self.query_one('#server_address', Input)
        self.app.set_server(host.value)  # pyright: ignore[reportAttributeAccessIssue]


class SourceSelectionView(VerticalGroup):
    def __init__(self, status):
        super().__init__()
        self.status = status

    def compose(self):
        source_is_disk = self.status.source == Source.Disk
        with Border('Select CL source (server/disk):'):
            with TabbedContent(initial='disk_vg' if source_is_disk else 'server_vg'):
                with TabPane('Disk', id='disk_vg'):
                    yield DiskOptions(self.status)
                with TabPane('Server', id='server_vg'):
                    yield ServerOptions(self.status)


class StatusView(Static):
    status = reactive(None, layout=True, init=False)

    def render(self):
        return self.status


class StatesView(VerticalScroll):
    strat_states = reactive({}, recompose=True)

    def __init__(self):
        VerticalScroll.__init__(self)
        bind(self, 'status', 'strat_states')

    def compose(self):
        if self.strat_states:
            for idx, strat in enumerate(self.strat_states.values()):
                pretty = maybe_else(
                    'N/A', lambda m: metamodel_rich_summary(m), strat.curr_meta_model
                )
                yield Border(strat.src_dir_abs_path, Static(pretty))


class StrategySelector(VerticalGroup):
    mmodel: reactive[None | MetaModel] = reactive(None, recompose=True)
    strat_states: reactive[dict[str, StrategyState]] = reactive({})

    def __init__(self):
        VerticalGroup.__init__(self)
        bind(self, 'strat_states', 'mmodel')

    def compose(self):
        strat_states = self.strat_states or {}
        kwargs = maybe_else({}, lambda m: {'value': m.src_dir_abs_path}, self.mmodel)

        if strat_states:
            with Border('Select an available strategy:'):
                yield Select(
                    options=map(dup, strat_states.keys()),
                    prompt='Click to select a strategy',
                    allow_blank=False,
                    **kwargs,
                )
        else:
            yield Static('No strategies available.')


class OverviewScreen(Screen):
    """Overview Screen"""

    DEFAULT_CSS = """OverviewScreen Border {
        border: round $primary;
        border-title-align: center;
        padding: 0 1 0 1;
    }"""

    status = reactive(Status(source=Source.Server))

    def __init__(self):
        Screen.__init__(self)
        self.title = 'Overview'
        bind(self, 'status')

    def compose(self):
        yield MyHeader()
        with VerticalGroup():
            with HorizontalGroup():
                yield SourceSelectionView(self.status)
                with VerticalGroup():
                    yield Border(
                        'TUI Status', StatusView().data_bind(OverviewScreen.status)
                    )
                    yield StrategySelector()
            yield Static(' [b]Summary of available strategies[/b]')
            yield StatesView()

        yield Footer()

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed):
        self.app.do_select_strategy(event.value)  # pyright: ignore[reportAttributeAccessIssue]



#
#   Imandra Inc.
#
#   codelogician/tools/filesystem.py
#

from __future__ import annotations

import fnmatch
from functools import partial
from pathlib import Path

import networkx as nx
import structlog
from pydantic import BaseModel

from .base import DepGraph, NodeMixin, nx_graph_from_adjacency_list

logger = structlog.get_logger(__name__)


class File(BaseModel, NodeMixin):
    """File entry with content and relative path"""

    relative_path: Path
    content: str

    def __hash__(self) -> int:
        return hash(self.relative_path)

    @property
    def path_key(self) -> str:
        return str(self.relative_path)

    def __repr__(self) -> str:
        parts = self.relative_path.parts
        if parts == ():
            return '.'
        return str(parts[-1])


class Dir(BaseModel, NodeMixin):
    """Directory entry with relative path"""

    relative_path: Path

    def __hash__(self) -> int:
        return hash(self.relative_path)

    @property
    def path_key(self) -> str:
        return str(self.relative_path)

    def __repr__(self) -> str:
        return str(self.relative_path)


FSEntry = File | Dir


class FileSystem(DepGraph[FSEntry]):
    def __init__(
        self,
        nx_graph: nx.DiGraph,
        repo_path: Path,
        file_extensions: list[str] | None = None,
        ignore_patterns: list[str] | None = None,
    ):
        super().__init__(nx_graph)

        if self.nx_graph.number_of_nodes() != 0 and not self.is_tree:
            raise ValueError('FileSystem must be a tree')

        self.repo_path = repo_path
        self.file_extensions = file_extensions
        self.ignore_patterns = ignore_patterns

    @staticmethod
    def from_disk(
        repo_path: Path,
        file_extensions: list[str] | None = None,
        respect_gitignore: bool = True,
    ) -> FileSystem:
        """Walk the file tree and return a FileSystem object

        Args:
            repo_path: The root path of the repository
            file_extensions: The file extensions to include in the file tree
            respect_gitignore: Whether to respect the gitignore file

        Returns:
            A FileSystem object with relative paths as nodes and edges indicating
            containment
        """

        if not (isinstance(repo_path, Path)):
            raise Exception(
                f'Attempting to read from disk but path is not a proper Path object: {repo_path}'
            )

        if file_extensions is None:
            file_extensions = []
        else:
            file_extensions = [ext.lstrip('.') for ext in file_extensions]

        if respect_gitignore:
            ignore_patterns = load_gitignore(repo_path)
            ignore_predicate = partial(is_ignored, patterns=ignore_patterns)
        else:
            ignore_patterns = []

            def ignore_predicate(path: Path) -> bool:
                return False

        data: dict[FSEntry, set[FSEntry]] = {}

        # Convert repo_path to relative path for consistent handling
        repo_path = repo_path.resolve()

        # Ensure root exists even if repo_path is empty
        root = Dir(relative_path=Path())
        data[root] = set()

        # Walk through all files and directories
        for entry in repo_path.rglob('*'):
            # Get relative path from repo root
            rel_path = entry.relative_to(repo_path)

            # Skip if ignored
            if ignore_predicate(rel_path):
                logger.debug('Ignoring entry by gitignore', entry=rel_path)
                continue

            # Filter by file extensions if specified
            if (
                file_extensions
                and entry.is_file()
                and not any(entry.suffix == f'.{ext}' for ext in file_extensions)
            ):
                logger.debug('Ignoring entry by file extension', entry=rel_path)
                continue

            # Create FSEntry object with content
            if entry.is_file():
                try:
                    content = entry.read_text(encoding='utf-8')
                    fs_entry = File(relative_path=rel_path, content=content)
                except Exception as e:
                    logger.warning(
                        'Failed to read file content', entry=rel_path, error=e
                    )
                    continue
            else:
                fs_entry = Dir(relative_path=rel_path)

            # Add entry to data structure
            if fs_entry not in data:
                data[fs_entry] = set()

            # Add parent-child relationship
            if rel_path != Path():  # Skip root
                parent_path = rel_path.parent
                parent_fs_entry = Dir(relative_path=parent_path)

                if parent_fs_entry not in data:
                    data[parent_fs_entry] = set()
                data[parent_fs_entry].add(fs_entry)

        # Unpack
        nodes = list(data.keys())
        adj_list = {
            node.path_key: {tgt.path_key for tgt in tgts} for node, tgts in data.items()
        }

        nx_graph = nx_graph_from_adjacency_list(nodes=nodes, adjacency_list=adj_list)

        return FileSystem(
            nx_graph=nx_graph,
            repo_path=repo_path,
            file_extensions=file_extensions,
            ignore_patterns=ignore_patterns,
        )


def load_gitignore(repo_path: Path) -> list[str]:
    """Load gitignore patterns from .gitignore file"""
    patterns = []
    gitignore_file = repo_path / '.gitignore'

    # Default patterns to always ignore
    default_patterns = [
        '.venv/',
        'venv/',
        '__pycache__/',
        '*.pyc',
        '*.pyo',
        '*.pyd',
        '.Python',
        'build/',
        'develop-eggs/',
        'dist/',
        'downloads/',
        'eggs/',
        '.eggs/',
        'lib/',
        'lib64/',
        'parts/',
        'sdist/',
        'var/',
        'wheels/',
        '*.egg-info/',
        '.git/',
        '.*',
    ]
    patterns.extend(default_patterns)

    if gitignore_file.exists():
        try:
            with gitignore_file.open('r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    if line and not line.startswith('#'):
                        patterns.append(line)
        except Exception:
            pass  # Ignore errors reading gitignore

    return patterns


def is_ignored(path: Path, patterns: list[str]) -> bool:
    """Check if a path matches any gitignore pattern"""
    path_str = str(path)
    path_parts = path.parts

    def matches(pattern):
        # Remove leading/trailing slashes for consistent matching
        pattern = pattern.strip('/')

        # Check if any part of the path matches the pattern
        # or if individual path parts
        return (
            fnmatch.fnmatch(path_str, pattern)
            or fnmatch.fnmatch(path_str, f'*/{pattern}')
            or fnmatch.fnmatch(path_str, f'{pattern}/*')
            or any(fnmatch.fnmatch(p, pattern) for p in path_parts)
        )

    return any(map(matches, patterns))


def experiment():
    fs = FileSystem.from_disk(Path('data/code4'))

    from codelogician.dep_tools.python import build_py_dep_graph

    py_dep = build_py_dep_graph(fs)
    print(py_dep.edges_repr())
    # two.py --> one.py

    print(fs.edges_repr())
    print(fs.graph_repr())


if __name__ == '__main__':
    experiment()



#
#   Imandra Inc.
#
#   cl_agent_state.py
#

from enum import StrEnum

from imandra.u.agents.code_logician.base import (
    VG,
    FormalizationStatus,
    RegionDecomp,
    TopLevelDefinition,
)
from imandra.u.agents.code_logician.graph import GraphState
from imandrax_api_models import Error
from pydantic import BaseModel, Field
from rich.text import Text

from ..modeling.task import ModelTask
from ..tools.cl_iml import ErrorSnippet

ErrorContext = list[ErrorSnippet]


class CLAgentState(BaseModel):
    """
    Wrapper around the CL agent state - this is what we get back from CL.
    """

    status: FormalizationStatus = FormalizationStatus.UNKNOWN
    src_code: str = ''
    iml_code: str | None = None  # IML code with artifacts (e.g. VGs)
    iml_model: str | None = None  # IML code without artifacts

    vgs: list[VG] = []
    region_decomps: list[RegionDecomp] = Field(default_factory=list[RegionDecomp])
    opaque_funcs: list[TopLevelDefinition] = Field(
        default_factory=list[TopLevelDefinition]
    )
    context: str = ''
    error_contexts: list[ErrorContext] = Field(default_factory=list[ErrorContext])
    errors: list[Error] = Field(default_factory=list[Error])

    eval_res: str = ''

    # These things are massive, so we don't save them to disk
    graph_state: GraphState | None = Field(default=None, exclude=True)

    def toJSON(self):
        """
        Return a dictionary we can save to disk
        """
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j: str | dict):
        """
        Create a CLAgentState value from JSON
        """
        if isinstance(j, str):
            return CLAgentState.model_validate_json(j)
        else:
            return CLAgentState.model_validate(j)

    def __repr__(self):
        """
        Return a nice set here
        """

        iml_code = str(self.iml_code) if self.iml_code else 'N\\A'
        iml_model = str(self.iml_model) if self.iml_model else 'N\\A'

        s = 'Agent state:\n'
        s += f'Status: {self.status}\n'
        s += f'Src Code: \n{self.src_code[:100]}\n'
        s += f'IML Code: \n{iml_code[:100]}\n'
        s += f'IML Model: \n{iml_model[:100]}\n'
        s += f'Context: \n{self.context}\n'
        s += f'Opaque funcs: \n{self.opaque_funcs}\n'
        s += f'Decomps: \n{self.region_decomps}\n'
        s += f'VGs: \n{self.vgs}\n'
        return s


def frm_status_to_rich(frm_status: FormalizationStatus):
    """
    Nicely format FormalizationStatus value
    """

    match frm_status:
        case FormalizationStatus.UNKNOWN:
            return Text(text='Unknown', style='bold red')
        case FormalizationStatus.INADMISSIBLE:
            return Text(text='Inadmissible', style='bold red')
        case FormalizationStatus.TRANSPARENT:
            return Text(text='Transparent', style='bold green3')
        case FormalizationStatus.EXECUTABLE_WITH_APPROXIMATION:
            return Text(text='Approximated', style='bold light green')
        case FormalizationStatus.ADMITTED_WITH_OPAQUENESS:
            return Text(text='Opaqueness', style='bold yellow')
        case _:
            raise Exception(f'Unknown code: {frm_status}')


class CLResultStatus(StrEnum):
    """
    CLResultStatus
    """

    SUCCESS = 'Success'
    ERROR = 'Error'
    TIMEOUT = 'Timeout'


class CLResult(BaseModel):
    """
    CLResult
    """

    task: ModelTask
    status: CLResultStatus
    agent_state: CLAgentState

    def __repr__(self):
        """Nice representation"""
        return f'CodeLogician result: [status={self.status.name}; taskID={self.task.task_id}]'



from typing import Dict
from .models import User, Project, Task


class InMemoryStorage:
    def __init__(self):
        self.users: Dict[int, User] = {}
        self.projects: Dict[int, Project] = {}
        self.tasks: Dict[int, Task] = {}

    def add_user(self, user: User):
        self.users[user.id] = user

    def add_project(self, project: Project):
        self.projects[project.id] = project

    def add_task(self, task: Task, project_id: int):
        self.tasks[task.id] = task
        if project_id in self.projects:
            self.projects[project_id].tasks.append(task)



#
#   Imandra Inc.
#
#   codelogician/commands/server/cmd_metamodel.py


from __future__ import annotations

import json
from typing import Any

import httpx
import typer

app = typer.Typer(
    name='metamodel',
    help='CLI for CodeLogician metamodel server endpoints.',
)
cache_app = typer.Typer(name='cache', help='Metamodel cache endpoints.')
app.add_typer(cache_app)


# -------------------------
# Helpers
# -------------------------


def _print_json(data: Any) -> None:
    typer.echo(json.dumps(data, indent=2, ensure_ascii=False))


def _handle_http_error(resp: httpx.Response) -> None:
    try:
        payload = resp.json()
    except Exception:
        payload = None

    if isinstance(payload, dict) and 'detail' in payload:
        raise typer.BadParameter(f'HTTP {resp.status_code}: {payload["detail"]}')
    raise typer.BadParameter(f'HTTP {resp.status_code}: {resp.text}')


def _get(
    base_url: str,
    path: str,
    *,
    params: dict[str, Any] | None = None,
    timeout: float,
) -> Any:
    url = base_url.rstrip('/') + path
    with httpx.Client(timeout=timeout) as client:
        resp = client.get(url, params=params)
    if resp.status_code >= 400:
        _handle_http_error(resp)
    return resp.json()


def _cfg(ctx: typer.Context) -> dict[str, Any]:
    """
    Read configuration from root Typer context.
    """
    root = ctx.find_root()
    assert isinstance(root.obj, dict)
    return root.obj


# -------------------------
# Commands
# -------------------------


@app.command('latest')
def metamodel_latest(ctx: typer.Context) -> None:
    """GET /metamodel — latest metamodel JSON."""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/metamodel', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(data)


@app.command('summary')
def metamodel_summary(ctx: typer.Context) -> None:
    """GET /metamodel/summary — summary JSON."""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/metamodel/summary', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(data)


@app.command('list')
def metamodel_list(
    ctx: typer.Context,
    listby: str = typer.Option(
        'alphabet',
        '--listby',
        help='Sort/group criteria: alphabet | frm_status | upstream | opaques | failed_vgs | ...',
    ),
) -> None:
    """GET /metamodel/list?listby=..."""
    cfg = _cfg(ctx)
    data = _get(
        cfg['base_url'],
        '/metamodel/list',
        params={'listby': listby},
        timeout=cfg['timeout'],
    )
    _print_json(data) if cfg['json_out'] else typer.echo(data)


@app.command('vgs')
def metamodel_vgs(ctx: typer.Context) -> None:
    """GET /metamodel/vgs — verification goals."""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/metamodel/vgs', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(data)


@app.command('decomps')
def metamodel_decomps(ctx: typer.Context) -> None:
    """GET /metamodel/decomps — decompositions."""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/metamodel/decomps', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(data)


@app.command('opaques')
def metamodel_opaques(ctx: typer.Context) -> None:
    """GET /metamodel/opaques — opaque functions."""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/metamodel/opaques', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(data)


# -------------------------
# Cache subcommands
# -------------------------


@cache_app.command('indices')
def cache_indices(ctx: typer.Context) -> None:
    """GET /metamodel/cache/indices"""
    cfg = _cfg(ctx)
    data = _get(cfg['base_url'], '/metamodel/cache/indices', timeout=cfg['timeout'])
    _print_json(data) if cfg['json_out'] else typer.echo(data)


@cache_app.command('get')
def cache_get(
    ctx: typer.Context,
    idx: int = typer.Argument(..., help='Cache index to fetch.'),
) -> None:
    """GET /metamodel/cache/by_index/{idx}"""
    cfg = _cfg(ctx)
    data = _get(
        cfg['base_url'],
        f'/metamodel/cache/by_index/{idx}',
        timeout=cfg['timeout'],
    )
    _print_json(data) if cfg['json_out'] else typer.echo(data)


if __name__ == '__main__':
    app()



#
#   Imandra Bank
#
#   bank.py
#

from account import Account
from asset import Asset
from currency import Currency
from ledger import Ledger
from liability import Liability
from transaction import Transaction, DebitTransaction, CreditTransaction

class Bank:
  """ Bank class to manage transactions """

  def __init__(self):
    """ Initialize the bank with an empty transaction list """
    self._ledger = Ledger() # balance sheet
    self._accounts = [] # list of customer accounts

  def get_transactions(self):
    """ Get all transactions in the bank """
    return self._transactions

  def create_account (self, customer: str, name: str, currency: int, account_type: int, account_number: str, starting_balance: float = 0.0):
    """ Create a new account in the bank """
    account = Account(customer, name, currency, account_type, account_number, starting_balance)
    self._accounts.append(account)
    return account
  
  def add_account (self, acc : Account):
    """ Add an already existing account object """
    self._accounts.append(acc)
  
  def get_account_by_id (self, acc_number):
    """ return account if it exists that matched the ID """

    acc = next (a for a in self._accounts if a.number() == acc_number)

    return acc

  def add_transaction(self, t: Transaction):
    """ Process a transaction and update the balance sheet """

    accOne = self.get_account_by_id ( t.from_account() )
    accTwo = self.get_account_by_id ( t.to_account()   )

    if accOne: accOne.apply_transaction(t)
    if accTwo: accTwo.apply_transaction(t)

  def __repr__(self):
    """ Return a nice representation of the bank """
    
    accountStr = "\n".join(map (str, self._accounts))
    bsStr = str(self._ledger)

    return f"Bank with Accounts: \n{accountStr}\n and Balance Sheet: \n {bsStr}"

if __name__ == "__main__":
    # Example usage
    
    bank = Bank()

    bank.add_account ( Account( "John Doe"  , "Savings Account"   , Currency.USD  , Account.ASSET   , "123456789"   , 1000.0 ) )
    bank.add_account ( Account( "John Doe"  , "Savings Account"   , Currency.USD  , Account.ASSET   , "123456789"   , 1000.0 ) )
    bank.add_account ( Account( "Jane Doe"  , "Checking Account"  , Currency.USD  , Account.ASSET   , "987654321"   , 1500.0 ) )
    bank.add_account ( Account( "John Doe"  , "Savings Account"   , Currency.USD  , Account.ASSET   , "123456789"   , 1000.0 ) )

    # let's now add some transactions
    bank.add_transaction(DebitTransaction("123456789", "987654321", 100))
    bank.add_transaction(CreditTransaction("987654321", "123456789", 50))

    print (bank)


#
# Imandra Inc.
#
# model_cmd_endpoints.py
#

import logging

from fastapi import HTTPException
from imandra.u.agents.code_logician.command import Command

from codelogician.modeling.model import Model
from codelogician.modeling.renderer import model_to_json
from codelogician.strategy.events import ModelCLTaskEvent

from ..cl_server import CLServer

log = logging.getLogger(__name__)


def register_model_endpoints(app: CLServer):  # noqa: C901
    """
    Individual model commands
    """

    @app.post('/model/cmd/freeze/{index}', operation_id='freeze_model')
    async def model_freeze(index: int):
        """
        Freeze the model
        """

        log.info(f'Submitting model freeze command for model with index={index}')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        models = list(state.curr_meta_model.models.values())

        if not (0 <= index and index < len(models)):
            raise HTTPException(
                status_code=400,
                detail=f'Invalid index (number of models is {len(models)}: {index})',
            )

        model: Model = models[index]

        model.iml_code_frozen = True

        return 'OK'

    @app.post('/model/cmd/unfreeze/{index}', operation_id='unfreeze_model')
    async def model_unfreeze(index: int):
        """
        Unfreeze the model
        """

        log.info(f'Submitting model unfreeze command for model with index={index}')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        models = list(state.curr_meta_model.models.values())

        if not (0 <= index and index < len(models)):
            raise HTTPException(
                status_code=400,
                detail=f'Invalid index (number of models is {len(models)}: {index})',
            )

        model: Model = models[index]

        model.iml_code_frozen = False

        return 'OK'

    @app.post('/model/cmd/{path}', operation_id='post_model_command')
    async def model_command(path: str, cmd: Command):
        """
        Submit model CodeLogician agent command
        """

        log.info(f'Submitting command={cmd} for model with path=[{path}]')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        if state.curr_meta_model is None:
            raise HTTPException(status_code=404, detail='No MetaModel exists!')

        if path not in state.curr_meta_model.models:
            raise HTTPException(
                status_code=404,
                detail=f'Could not locate model with specified path=[{path}]',
            )

        try:
            taskEvent = ModelCLTaskEvent(rel_path=path, cmd=cmd)
            app.strategy_worker_by_id(strategy_id).add_event(taskEvent)
        except Exception as e:
            errMsg = f'Failed to create a task for CodeLogician: {str(e)}'
            log.error(errMsg)
            raise HTTPException(status_code=403, detail=errMsg)

        return 'OK'

    @app.get('/model/paths', operation_id='get_model_paths')
    async def model_paths():
        """
        Get list of all paths for which models exist.
        """

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        log.info('Request for model paths')
        if state.curr_meta_model is None:
            return []
        else:
            return list(state.curr_meta_model.models.keys())

    @app.get('/model/byindex/{index}')
    async def model_by_index(index: int):
        """
        Retrieve a JSON represention of a Model for specified index (by model rel paths)
        """

        log.info(f'Request for model by index:{index}')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        models = list(state.curr_meta_model.models.values())

        if not (0 <= index and index < len(models)):
            raise HTTPException(
                status_code=400,
                detail=f'Invalid index (number of models is {len(models)}: {index})',
            )

        return models[index].model_dump_json()

    @app.get('/model/bypath/{path}', operation_id='get_model_by_path')
    async def model_by_path(path: str):
        """
        Retrieve JSON representation of a model for a specified path (which is relative to the source directory)
        """

        log.info(f'Request for model by path: {path}')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        if path in state.curr_meta_model.models:
            return model_to_json(state.curr_meta_model.models[path])
        else:
            raise HTTPException(
                status_code=404, detail=f'Model with specific path [{path}] not found!'
            )



#
#   Imandra Inc.
#
#   codelogician/modeling/task.py
#

import datetime
import uuid
from enum import StrEnum

from imandra.u.agents.code_logician.base import FormalizationDependency
from imandra.u.agents.code_logician.command import (
    AgentFormalizerCommand,
    Command,
    EditStateElementCommand,
    GenFormalizationDataCommand,
    GenModelCommand,
    GenRegionDecompsCommand,
    GenVgsCommand,
    InitStateCommand,
    InjectFormalizationContextCommand,
    SetModelCommand,
)
from imandra.u.agents.code_logician.graph import GraphState
from pydantic import BaseModel, Field


class ModelTaskMode(StrEnum):
    """Strategy mode"""

    AGENT = 'Agent'
    MANUAL = 'Manual'


class TaskKind(StrEnum):
    USER_IML_UPDATE = 'USER_IML_UPDATE'
    FORMALIZE = 'FORMALIZE'
    ANALYZE = 'ANALYZE'


class FormalizationNeed(StrEnum):
    """
    Reasons to perform formalization
    """

    # fmt: off
    NO_AGENT_STATE    = 'No_agent_state'   # This is the first time we're formalizing the model
    SRC_CODE_CHANGED  = 'Src_code_changed' # Source code has changed
    IML_CODE_CHANGED  = 'IML_code_changed' # User edited the IML code, so we need to resend it to CL
    CONTEXT_ADDED     = 'Cntxt_added'      # Human feedback has been provided
    DEPS_CHANGED      = 'Deps_changed'     # Dependencies (at least one) have changed
    # fmt: on


class ModelTask(BaseModel):
    """
    Model task contains a single task for CodeLogician to execute
    """

    # fmt: off
    task_kind        : TaskKind = TaskKind.FORMALIZE
    rel_path         : str
    src_code         : str = ''
    context          : str | None = ''
    dependencies     : list[FormalizationDependency] = Field(default_factory=list)
    graph_state      : GraphState | None = None
    language         : str = 'Python'
    mode             : ModelTaskMode = ModelTaskMode.AGENT
    gen_vgs          : bool = False
    gen_decomps      : bool = False
    # fmt: on

    # If set, we disregard everything else and just return these
    # along with the existing state.
    specified_commands: list[Command] | None = None

    task_id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now)

    def is_empty(self) -> bool:
        return self.src_code.strip() == ''

    def commands(self) -> list[Command]:
        """
        Return a list of CodeLogician agent commands for this task.
        """
        if self.specified_commands is not None:
            return self.specified_commands

        cmds: list[Command] = []

        # Always init from current src
        cmds.append(InitStateCommand(src_code=self.src_code, src_lang=self.language))

        # Optional context
        if self.context is not None:
            cmds.append(InjectFormalizationContextCommand(context=self.context))

        # Optional deps
        if self.dependencies:
            cmds.append(
                EditStateElementCommand(update={'dependency': self.dependencies})
            )

        # Mode
        if self.mode == ModelTaskMode.MANUAL:
            cmds.append(GenFormalizationDataCommand())
            cmds.append(GenModelCommand())
        elif self.mode == ModelTaskMode.AGENT:
            cmds.append(
                AgentFormalizerCommand(
                    no_gen_model_hitl=True,
                    max_tries_wo_hitl=3,
                    max_tries=3,
                    no_check_formalization_hitl=True,
                    no_refactor=False,
                )
            )
        else:
            raise Exception(f'Attempting unrecognized mode: {self.mode}')

        # Optional artifacts
        if self.gen_vgs:
            cmds.append(GenVgsCommand(description=''))

        if self.gen_decomps:
            cmds.append(GenRegionDecompsCommand(function_name=None))

        return cmds

    def toJSON(self) -> str:
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j: dict | str) -> 'ModelTask':
        if isinstance(j, str):
            return ModelTask.model_validate_json(j)
        return ModelTask.model_validate(j)

    def __repr__(self) -> str:
        return f'Base ModelTask with ID = {self.task_id}; path={self.rel_path}'


class DecompVGsModelTask(ModelTask):
    """
    Here we just perform extraction/running of Decomp/VGs
    """

    task_kind: TaskKind = TaskKind.ANALYZE

    def commands(self) -> list[Command]:
        if self.specified_commands is not None:
            return self.specified_commands

        if self.graph_state is None:
            raise ValueError(
                'DecompVGsModelTask requires graph_state; run formalization first.'
            )

        cmds: list[Command] = []
        if self.context is not None:
            cmds.append(InjectFormalizationContextCommand(context=self.context))
        if self.dependencies:
            cmds.append(
                EditStateElementCommand(update={'dependency': self.dependencies})
            )

        if self.gen_vgs:
            cmds.append(GenVgsCommand(description=''))
        if self.gen_decomps:
            cmds.append(GenRegionDecompsCommand(function_name=None))

        return cmds

    def __repr__(self) -> str:
        return f'DecompVGsModelTask with ID = {self.task_id}; path={self.rel_path}'


class UserManualIMLEditTask(ModelTask):
    """
    This updates the graph state from a manual user IML edit
    """

    task_kind: TaskKind = TaskKind.USER_IML_UPDATE

    iml_code: str  # user provided IML code

    def is_empty(self) -> bool:
        return False

    def commands(self) -> list[Command]:
        """
        We're just going to invoke the commands.
        """
        if self.graph_state:
            return [SetModelCommand(model=self.iml_code)]

        return [
            InitStateCommand(src_code=self.src_code, src_lang=self.language),
            SetModelCommand(model=self.iml_code),
        ]

    def __repr__(self) -> str:
        return f'UserManualIMLEditTask with ID = {self.task_id}; path={self.rel_path}'



#
#   Imandra Inc.
#
#   intro.py
#

from pathlib import Path

from textual.screen import Screen
from textual.widgets import Footer, Label
from textual_image.widget import HalfcellImage as Image


def local_file(name):
    return Path(__file__).parent / name


class IntroScreen(Screen):
    def on_mount(self):
        self.styles.align_vertical = 'middle'
        self.styles.align_horizontal = 'center'
        self.styles.background = 'black'

    def compose(self):
        banner = Image(local_file('../data/splash.png'))
        banner.styles.padding = (0, 0, 0, 0)
        banner.styles.width = 99
        banner.styles.height = 9
        yield banner
        yield Label('[$primary]Imandra CodeLogician, version 1.0')
        yield Footer()


if __name__ == '__main__':
    pass



LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'standard': {'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'},
        'custom_formatter': {
            'format': '%(asctime)s [%(processName)s: %(process)d] [%(threadName)s: %(thread)d] [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
        'stream_handler': {
            'formatter': 'custom_formatter',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
        'file_handler': {
            'formatter': 'custom_formatter',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': 'app.log',
            'maxBytes': 1024 * 1024 * 1,  # = 1MB
            'backupCount': 3,
        },
    },
    'loggers': {
        'uvicorn': {
            'handlers': ['default', 'file_handler'],
            'level': 'TRACE',
            'propagate': False,
        },
        'uvicorn.access': {
            'handlers': ['stream_handler', 'file_handler'],
            'level': 'TRACE',
            'propagate': False,
        },
        'uvicorn.error': {
            'handlers': ['stream_handler', 'file_handler'],
            'level': 'TRACE',
            'propagate': False,
        },
        'uvicorn.asgi': {
            'handlers': ['stream_handler', 'file_handler'],
            'level': 'TRACE',
            'propagate': False,
        },
    },
}
